/*
 * Created on 11/12/2004
 *  
 */
package main.aspects;

import logic.formulas.CompositeFormula;
import logic.formulas.FormulaFactory;
import logic.signedFormulas.SignedFormulaList;
import logicalSystems.classicalLogic.ClassicalRules;
import main.facade.TableauFacade;
import main.proofTree.Origin;
import main.proofTree.ProofTree;
import main.proofTree.SignedFormulaNode;
import main.strategy.ClassicalProofTree;
import main.strategy.Strategy;
import main.tableau.Execution;
import main.tableau.Proof;
import main.tableau.RulesUsage;
import rules.Rule;

/**
 * Aspect that counts the number of applications of rules in a proof.
 * 
 * @author Adolfo Gustavo Serra Seca Neto
 *  
 */
public aspect Profiler {

	
	private static boolean enabled = false;

	/**
	 * Habilita ou desabilitia todos os advices do aspecto 
	 *
	 */
	public static void setEnabled(boolean value) {
		enabled = value;
	}

	int timesSimple = 0;

	int timesSubformula = 0;

	int timesPB = 0;

	long initialMemory, maxMemoryUsed = 0;

	int lengthBefore, lengthAfter;

	RulesUsage rulesUsage;

	FormulaFactory _lastFF;

	public ProofTree TableauFacade._lastPT;

	ProofTree _lastPT;

	TableauFacade _facade;

	// pointcuts
	pointcut callClose():execution (* Strategy+.close(..));

	pointcut callSimpleRule(Rule r, SignedFormulaList sfl):
        target(r) && args(..,sfl) &&
        call (* Rule+.getPossibleConclusions(..));

	pointcut callSubformulaRule(Rule r, SignedFormulaList sfl,
			CompositeFormula f):
        target(r) && args(..,sfl, f) &&
        call (* Rule+.getPossibleConclusions(..));

	pointcut callPBRule():
        call (private * Strategy+.applyPBOnce(..));

	pointcut callRule_toGetFormulaFactory(FormulaFactory ff):
        args(*,ff, ..) &&
        call (* Rule+.getPossibleConclusions(..));

	pointcut setExecution(Execution e):
        args(e) && 
        call(* Proof.setExecution(*));

	pointcut addNodeToProofTree(ProofTree pt, SignedFormulaNode sfn):
        target(pt) && args (sfn) && (call (* ProofTree+.addLast(*)) ||
        call (ProofTree+.new(*)));

	pointcut getFacade(TableauFacade tf):
        target(tf) &&  (execution (TableauFacade.new(..))) ;

	pointcut TableauFacadeMain():
        (execution (public static void TableauFacade.main(..)));

	pointcut TableauFacadeProve():
        (execution (* TableauFacade.prove(..)));

	pointcut ProofNew():
        (execution (Proof.new(..)));

	pointcut OutOfMemory():
        handler (OutOfMemoryError);

	pointcut ClassicalProofTree(ClassicalProofTree cpt):
        target(cpt) &&
        (execution  (* ClassicalProofTree+.addLast(*)));

	after(ClassicalProofTree cpt): ClassicalProofTree(cpt){
		//        System.out.println("Size of proof tree: " + cpt.sizeOf());
	}

	//    before(): OutOfMemory(){
	//        System.err.println("Out of memory...");
	//        System.out.println(thisJoinPoint.toString());
	//        System.out.println(thisJoinPoint.getThis().getClass());
	//    }

	// pieces of advice
	before(): callClose(){

		if (enabled) {
			rulesUsage = new RulesUsage();

			timesSimple = timesSubformula = timesPB = 0;
			System.gc();
			initialMemory = Runtime.getRuntime().freeMemory();
		}
	}

	after(Execution e): setExecution(e){
		if (enabled) {
			e.setRulesUsage(rulesUsage);
		}
	}

	SignedFormulaList around(Rule r, SignedFormulaList sfl): 
        callSimpleRule(r, sfl){
		if (enabled) {

			try {
				checkMemory(timesSimple);
				timesSimple++;

				rulesUsage.addToUsage(r);
				rulesUsage.addToUsage(RulesUsage.SIMPLE);

				SignedFormulaList sflResult = proceed(r, sfl);
				return sflResult;
			} catch (OutOfMemoryError e) {
				System.err
						.println("Memory finished after calling simple rule...");
				exitGracefully();
			}

			return null;
		} else {
			return proceed(r, sfl);
		}
	}

	// see around before
	//    before(Rule r, SignedFormulaList sfl):
	//        callSimpleRule(r, sfl) {
	//        checkMemory(timesSimple);
	//        timesSimple++;
	//
	//        rulesUsage.addToUsage(r);
	//        rulesUsage.addToUsage(RulesUsage.SIMPLE);
	//    }
	//
	//    after(Rule r, SignedFormulaList sfl):
	//        callSimpleRule(r, sfl) {
	//        checkMemory(timesSimple);
	//    }

	SignedFormulaList around(Rule r, SignedFormulaList sfl, CompositeFormula f): 
        callSubformulaRule(r, sfl,f) {
		if (enabled) {
			try {
				checkMemory(timesSubformula);
				timesSubformula++;

				rulesUsage.addToUsage(r);
				rulesUsage.addToUsage(RulesUsage.SUBFORMULA);

				SignedFormulaList sflResult = proceed(r, sfl, f);
				return sflResult;
			} catch (OutOfMemoryError e) {
				System.err
						.println("Memory finished after calling subformula rule...");
				exitGracefully();
			}

			return null;
		} else {
			return proceed(r, sfl, f);
		}

	}

	// see around before
	//    before(Rule r, SignedFormulaList sfl, CompositeFormula f):
	//        callSubformulaRule(r, sfl,f) {
	//        checkMemory(timesSubformula);
	//        timesSubformula++;
	//
	//        rulesUsage.addToUsage(r);
	//        rulesUsage.addToUsage(RulesUsage.SUBFORMULA);
	//    }
	//
	//    after(Rule r, SignedFormulaList sfl, CompositeFormula f):
	//            callSubformulaRule(r, sfl,f) {
	//        checkMemory(timesSubformula);
	//
	//    }

	boolean around(): 
        callPBRule() {

		if (enabled) {
			try {
				checkMemory(timesPB);
				timesPB++;

				rulesUsage.addToUsage(ClassicalRules.PB);
				rulesUsage.addToUsage(RulesUsage.PB);

				boolean result = proceed();
				return result;
			} catch (OutOfMemoryError e) {
				System.err.println("Memory finished after calling PB rule...");
				exitGracefully();
			}

			return false;
		} else
			return proceed();

	}

	// see around before
	//    before():
	//        callPBRule() {
	//
	//        checkMemory(timesPB);
	//        timesPB++;
	//
	//        rulesUsage.addToUsage(ClassicalRules.PB);
	//        rulesUsage.addToUsage(RulesUsage.PB);
	//    }
	//
	//    after():
	//            callPBRule() {
	//        checkMemory(timesSubformula);
	//
	//    }

	void around(ProofTree pt, SignedFormulaNode sfn): addNodeToProofTree(pt, sfn){
		if (enabled) {
			checkMemory(timesSubformula);
			try {
				lengthBefore = pt.getNumberOfNodes();
				_lastPT = pt;

				proceed(pt, sfn);
			} catch (OutOfMemoryError e) {
				System.err
						.println("Memory finished after adding node to proof tree...");
				exitGracefully();
				;
			}

			checkMemory(timesSubformula);
			lengthAfter = pt.getNumberOfNodes();

			if (lengthBefore == lengthAfter) {
				//            System.err.println("Nothing added..." + lengthAfter + " " +
				// sfn);
				if ((sfn.getOrigin() != null)
						&& (sfn.getOrigin() != Origin.PROBLEM)) {
					rulesUsage.addToUsageRepeated(sfn.getOrigin().getRule());
				}
			}
		} else
			proceed(pt, sfn);
	}

	//    before(ProofTree pt, SignedFormulaNode sfn): addNodeToProofTree(pt, sfn){
	//        checkMemory(timesSubformula);
	//        lengthBefore = pt.getNodes();
	//        _lastPT = pt;
	//    }
	//
	//    after(ProofTree pt, SignedFormulaNode sfn): addNodeToProofTree(pt, sfn){
	//        checkMemory(timesSubformula);
	//        lengthAfter = pt.getNodes();
	//
	//        if (lengthBefore == lengthAfter) {
	//            // System.err.println("Nothing added..." + lengthAfter + " " + sfn);
	//            if (sfn.getOrigin() != Origin.PROBLEM) {
	//                rulesUsage.addToUsageRepeated(sfn.getOrigin().getRule());
	//            }
	//        }
	//
	//    }

	after(FormulaFactory ff): callRule_toGetFormulaFactory(ff){
		if (enabled) {
			_lastFF = ff;
		}

	}

	before(TableauFacade tf): getFacade(tf){
		if (enabled) {
			_facade = tf;
		}
	}

	before(): TableauFacadeMain(){
		if (enabled) {
			showFreeMemory("Before execution of TableauFacade's main: ");
		}
	}

	before(): TableauFacadeProve(){
		if (enabled) {
			showFreeMemory("Before execution of TableauFacade's prove: ");
		}
	}

	before(): ProofNew(){
		if (enabled) {
			showFreeMemory("Before execution of Proof's new: ");
		}
	}

	after(): TableauFacadeProve(){
		if (enabled) {
			showFreeMemory("After execution of TableauFacade's prove: ");
		}
	}

	// methods of the aspect
	private void checkMemory(int times) {

		if (Runtime.getRuntime().freeMemory() < 50000) {

			System.out.println("Memory too low...");
			System.out.println("Memory available before garbage collection "
					+ Runtime.getRuntime().freeMemory());
			showFormulaFactorySize();
			System.out.println("Performing garbage collection ...");
			Runtime.getRuntime().gc();
			showFormulaFactorySize();
			showProofTreeSize();

			if (Runtime.getRuntime().freeMemory() < 100000) {

				System.out.println("Memory available after garbage collection "
						+ Runtime.getRuntime().freeMemory());
				System.out
						.println("It was not possible to free enough memory to continue. Exiting...");

				showRules();
				exitGracefully();
			}

		}

	}

	private void showRules() {
		//        System.out.println("Simple rules verified: " + timesSimple);
		//        System.out.println("Subformula rules verified: " + timesSubformula);
		//        System.out.println("PB Rules applied: " + timesPB);
		//        System.err.println(XMLViewer.getXMLElementAsString(rulesUsage.asXMLElement()));
	}

	private void showFormulaFactorySize() {
		System.out.println("Measurement of Formula Factory : "
				+ _lastFF.getComplexity());
		System.out.println("Number of atomic formulas in Formula Factory : "
				+ _lastFF.getAtomicFormulasSize());
		System.out.println("Number of composite formulas in Formula Factory : "
				+ _lastFF.getCompositeFormulasSize());
		System.out
				.println("Measurement of complexity of the Formula Factory : "
						+ _lastFF.getComplexity());
		System.out.println("Size of Formula Factory structures: "
				+ _lastFF.getStructuresSize());
	}

	private void showProofTreeSize() {
		while (_lastPT.getParent() != null) {
			_lastPT = _lastPT.getParent();
		}
		//        System.out.println("Size of the proof tree- \n" +
		// _lastPT.showSize());
		_facade._lastPT = _lastPT;

	}

	private void exitGracefully() {

		System.out
				.println("------------------------ It is going to exit ... ------------------------");

		System.out
				.println("\nMemory available before garbage collection and liberating proof tree: "
						+ Runtime.getRuntime().freeMemory());

		Runtime.getRuntime().gc();
		System.out
				.println("\nMemory available after garbage collection and before liberating proof tree: "
						+ Runtime.getRuntime().freeMemory());

		showFormulaFactorySize();
		showProofTreeSize();
		_facade._lastPT = null;
		_facade = null;

		_lastPT = null;

		Runtime.getRuntime().gc();
		System.out
				.println("\nMemory available after second garbage collection and after liberating proof tree: "
						+ Runtime.getRuntime().freeMemory());

		//        System.exit(0);

		throw new OutOfMemoryError();

	}

	private void showFreeMemory(String message) {
		Runtime.getRuntime().gc();
		//        System.out
		//                .println("*** " + message + Runtime.getRuntime().freeMemory());
	}
}