/*
 * Created on 10/03/2005
 *  
 */
package main.aspects;

import java.io.FileWriter;

import main.facade.DirectoryStructure;
import main.facade.TableauFacade;
import main.tableau.Proof;

/**
 * An aspect for helping the tableau facade: save summary files
 * 
 * @author Adolfo Gustavo Serra Seca neto
 * 
 */
public aspect TableauFacadeAspect {

    // pointcuts for proving/analysing sequences

    pointcut proveSequence(TableauFacade tf):
        target(tf) &&
        call(* TableauFacade.proveSequence(..));

    pointcut proveInstanceInASequence(TableauFacade tf):
        execution (Proof TableauFacade.prove(..)) 
        &&
        cflow(proveSequence(tf));

    pointcut analyseSequence(TableauFacade tf):
        target(tf) &&
        call(* TableauFacade.analyseSequence(..));

    pointcut analyseInstanceInASequence(TableauFacade tf):
        execution (String TableauFacade.analyseProblem(..))  
        &&
        cflow(analyseSequence(tf));

    pointcut analyseIsolatedInstance(TableauFacade tf, String problemName):
        target(tf) &&
        execution (String TableauFacade.analyseProblem(*)) &&
        args(problemName);

    // before and after - open and close file
    before(TableauFacade tf): proveSequence(tf){
        if (tf.isSaveSummary()) {
            openSummariesFile(tf, RESULTS);
        }
    }

    after(TableauFacade tf): proveSequence(tf){
        if (tf.isSaveSummary()) {
            closeSummariesFile(tf, RESULTS);
        }
    }

    before(TableauFacade tf): analyseSequence(tf){
        if (tf.isSaveSummary()) {
            openSummariesFile(tf, ANALYSIS);
        }
    }

    after(TableauFacade tf): analyseSequence(tf){
        if (tf.isSaveSummary()) {
            closeSummariesFile(tf, ANALYSIS);
        }
    }

    // around - write in summary file
    Proof around(TableauFacade tf): proveInstanceInASequence(tf){
        Proof p = proceed(tf);
        if (tf.isSaveSummary()) {
            writeInstanceSummary(tf, p.summary(), RESULTS);
        }

        return p;
    }

    String around(TableauFacade tf): analyseInstanceInASequence(tf){
        String s = proceed(tf);
        if (tf.isSaveSummary()) {
            writeInstanceSummary(tf, s, ANALYSIS);
        }

        return s;
    }

    String around(TableauFacade tf, String problemName): analyseIsolatedInstance(tf, problemName){
        String s = proceed(tf, problemName);

        FileWriter summaryFile;
        if (tf.isSaveSummary()) {
            try {
                summaryFile = new FileWriter(generateSummariesFileName(tf,
                        problemName));
                summaryFile.write(s);
                summaryFile.close();
            } catch (Exception e) {
                System.err.println("Problem tring to write "
                        + generateSummariesFileName(tf, problemName));
            }
        }

        return s;
    }

    private FileWriter _summariesFile = null;

    public static final int RESULTS = 0;

    public static final int ANALYSIS = 1;

    private String generateSummariesFileName(TableauFacade tf,
            String problemName) {
        return tf.getSummariesDir() + problemName + tf.getSummaryTermination();
    }

    public void openSummariesFile(TableauFacade tf, int option) {
        try {
            _summariesFile = new FileWriter(generateSummariesFileName(tf,
                    option));
        } catch (Exception e) {
            System.err
                    .println("Unable to open for creation the following file: "
                            + generateSummariesFileName(tf, option));
        }
    }

    public void closeSummariesFile(TableauFacade tf, int option) {
        try {
            _summariesFile.close();
        } catch (Exception e) {
            System.err.println("Unable to close the following file: "
                    + generateSummariesFileName(tf, option));
        }
    }

    public void writeInstanceSummary(TableauFacade tf, String s, int option) {
        try {
            _summariesFile.write(s);
            _summariesFile.flush();
        } catch (Exception e) {
            System.err.println("Unable to write to the following file: "
                    + generateSummariesFileName(tf, option));
        }
    }

    private String generateSummariesFileName(TableauFacade tf, int terminationOption) {
        return tf.getDirectoryStructure().getCompletePath(DirectoryStructure.RESULTS,
                DirectoryStructure.SUMMARIES, tf.getClassOfProblems())
                + tf.getProblemName()
                + (((terminationOption == RESULTS) ? tf.getResultsTermination()
                        : tf.getAnalysisTermination())  + tf.getSummaryTermination());
    }
}