/*
 * Created on 10/03/2005
 *  
 */
package main.aspects;

import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.text.NumberFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Locale;
import java.util.Map;

import logic.problem.Problem;
import logic.problem.ProblemType;
import logic.signedFormulas.SignedFormula;
import logic.signedFormulas.SignedFormulaBuilder;
import logic.signedFormulas.SignedFormulaList;
import main.facade.TableauFacade;
import main.proofTree.Node;
import main.proofTree.ProofTree;
import main.tableau.Execution;
import main.tableau.Proof;
import main.tableau.RulesUsage;

import org.jdom.Document;
import org.jdom.ProcessingInstruction;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import util.MemoryManager;

/**
 * An aspect for profiling the tableau facade.
 * 
 * @author Adolfo Gustavo Serra Seca neto
 * 
 */
public aspect TableauFacadeProfilerAspect {

    private static final int NO_SIZE = -1;

    pointcut createProof(TableauFacade tf, Problem problem):
        this(tf) &&
        args(.., problem) &&
        call(Proof.new(..));

    /**
     * What should be done here?
     * 
     * Prove not only one time, but as many times as asked.
     * 
     * Before and after each proof, measure time spent.
     * 
     * After all proofs of the same problem, calculate average time spent.
     * 
     * 
     * 
     */

    /** do some profiling before and after trying to close a tableau */
    Proof around(TableauFacade tf, Problem problem): createProof(tf, problem){
        /** measures the time spent producing a proof */
        
        long total = 0;
        Proof p = null;

        /** tries to close a problem one or more times */
        for (int i = 0; i < tf.getNumberOfTimes(); i++) {
            long beginning, end;
            try {
                MemoryManager.runGC();
            } catch (Exception e) {
                // do nothing
            }
            
            /** gets time before proving */
            beginning = System.currentTimeMillis();

            
            try {
                /** proves the problem */
                p = proceed(tf, problem);
            } catch (OutOfMemoryError e) {
//                System.err.println(e.getMessage());
                throw e;
            }

            /** gets time after proving */
            end = System.currentTimeMillis();
            total += end - beginning;
        }
        return afterAroundCreateProof(tf, problem, p, total
                / tf.getNumberOfTimes());
    }

    /** do some profiling and returns the proof */
    private Proof afterAroundCreateProof(TableauFacade tf, Problem problem,
            Proof p, long timeSpent) {
        

        NumberFormat nf = NumberFormat.getInstance(Locale.US);
        nf.setMinimumFractionDigits(3);
        
        if (tf.isSaveSummary() ){
            p.setSummary(generateProofSummary(tf, timeSpent, nf, p, problem));
        }

        recordsInfoExecution(tf, problem, p, timeSpent);
        saveProof(tf, p);
        saveSimplifiedProof(tf, p);
        showSummary(tf, p, timeSpent, nf);

        return p;
    }

    /** records info about the execution in the proof */
    private void recordsInfoExecution (TableauFacade tf, Problem problem, Proof p, long timeSpent){ 
        if (tf.isSaveInfoExecution()){
    Execution e = new Execution(getHostName(), new Date());

    e
            .setOS(System.getProperty("os.name"), System
                    .getProperty("os.version"));
    e.setMemory(Runtime.getRuntime().maxMemory(), Runtime.getRuntime()
            .freeMemory(), Runtime.getRuntime().totalMemory());

    e.setFilename(_currentProblemFileName);
    e.setStrategy(tf.getProver().getStrategy());

    RulesUsage ru = new RulesUsage();
    e.setRulesUsage(ru);
    e.setTimeElapsed((timeSpent) / 1000.0);
    
    e.setSignedFormulaBuilder(new SignedFormulaBuilder(problem.getSignedFormulaFactory(), problem.getFormulaFactory()));

    p.setExecution(e);
	}
    }


    /** saves proof in a file */
    private void saveProof(TableauFacade tf, Proof p) {
        if (tf.isSaveProof()) {
            if (_currentProblemSize != NO_SIZE) {
                recordInFile(tf, _currentProblemName, _currentProblemSize,
                        _currentProblemBase, _currentProblemType, p);
            } else {
                recordInFile(tf, _currentProblemName, p);
            }
        }
    }

    /** saves simplified version of proof in a file */
    private void saveSimplifiedProof(TableauFacade tf, Proof p) {
        if (tf.isSaveSimpleProof()) {
            if (_currentProblemSize != NO_SIZE) {
                recordSimpleProofInFile(tf, _currentProblemName,
                        _currentProblemSize, _currentProblemBase,
                        _currentProblemType, p);
            } else {
                recordSimpleProofInFile(tf, _currentProblemName, p);
            }
        }
    }

    /** shows summary of proof on System.out */
    private void showSummary(TableauFacade tf, Proof p, long timeSpent, NumberFormat nf) {

        if (tf.isShowSummary()) {
            System.out.println(_currentProblemInstanceName + " closed? "
                    + p.isClosed() + " in " + nf.format((timeSpent) / 1000.0)
                    + " s");
        }
    }

    // TODO create a class 
    /** some info about the problem */
    public static String _currentProblemName, _currentProblemInstanceName,
            _currentProblemFileName, _currentProblemSummaryString;

    int _currentProblemSize, _currentProblemBase;

    ProblemType _currentProblemType;

    pointcut problemInfo(TableauFacade tf, String familyName, int size,
            int base, ProblemType type):
        target(tf) &&
        execution (* TableauFacade.prove(..)) &&
        args(familyName, size, base, type);

    // TODO instantiate a class
    /**
     * records info about problem in aspect attributes - for families of
     * problems
     */
    before(TableauFacade tf, String familyName, int size, int base,
            ProblemType type):  
        problemInfo(tf, familyName, size, base, type){
        _currentProblemName = familyName;
        _currentProblemInstanceName = tf.generateFullProblemName(familyName,
                size, base, type);
        _currentProblemFileName = tf.generateFileName(familyName, size, base,
                type);
        _currentProblemSummaryString = familyName + tf.getTexSeparator() + size;
        _currentProblemSize = size;
        _currentProblemBase = base;
        _currentProblemType = type;
    }

    pointcut problemInfoWithoutSize(TableauFacade tf, String problemName):
        target(tf) &&
        execution (* TableauFacade.prove(..)) &&
        args(problemName);

    /** records info about problem in aspect attributes - for isolated problems */
    before(TableauFacade tf, String problemName):  
        problemInfoWithoutSize(tf, problemName){
        _currentProblemName = problemName;
        _currentProblemInstanceName = tf.generateFullProblemName(problemName);
        _currentProblemFileName = tf.generateFileName(problemName);
        _currentProblemSummaryString = problemName;
        _currentProblemSize = NO_SIZE;
    }

    /** generates a string containing the summary of a proof */
    private String generateProofSummary(TableauFacade tf, long timeSpent,
            NumberFormat nf, Proof p, Problem problem) {
        String result = "";
        // . simplified proof tree length in bytes
        // . time spent
        // . number of nodes
        // . height
        // . formula factory complexity
        // . signed formula factory complexity
        // . size of problem
        // . size of proof tree
        result += _currentProblemSummaryString + tf.getTexSeparator();
        result += p.getProofTree().toStringShort().length()
                + tf.getTexSeparator();
        result += nf.format((timeSpent) / 1000.0) + tf.getTexSeparator();
        result += p.getProofTree().getNumberOfNodes() + tf.getTexSeparator();
        result += p.getProofTree().getHeight() + tf.getTexSeparator();
        result += problem.getFormulaFactory().getComplexity()
                + tf.getTexSeparator();
        result += problem.getSignedFormulaFactory().getComplexity()
                + tf.getTexSeparator();
        result += getSize(problem.getFormulas()) + tf.getTexSeparator();
        result += p.getProofTree().getSize();

        return result + tf.getTexFinalizer() + TableauFacade.LS;
    }

    // TODO Why an inter-type declaration?
    /**
     * @return the size of a proof tree
     */
    public long ProofTree.getSize() {

        long result = 0;

        Node current = getRoot();

        do {

            result = result
                    + ((SignedFormula) current.getContent()).getComplexity();
            current = current.getNext();

        } while (current != null);

        if (getLeft() != null) {
            result = result + getLeft().getSize() + getRight().getSize();
        }

        return result;
    }

    /**
     * @return the size of a signed formula list
     */
    public static long getSize(SignedFormulaList sfl) {

        Iterator it = sfl.iterator();

        long result = 0;

        while (it.hasNext()) {
            result += +((SignedFormula) it.next()).getComplexity();
        }

        return result;

    }

    private String getHostName() {
        try {
            Process p = Runtime.getRuntime().exec("hostname");
            InputStream os = p.getInputStream();
            try {
                p.waitFor();
            } catch (InterruptedException e1) {
                System.err.println("Error trying to get host name.");
            }
            String result = "";
            while (os.available() > 0) {
                String c = "" + (char) os.read();
                if (!c.equals(System.getProperty("line.separator"))) {
                    result += c;
                }
            }
            return result;

        } catch (IOException e) {
            System.err.println("Error trying to get host name.");
        }

        return "Error";
    }

    // TODO Profile objects should do this
    /** Recording XML style proofs */
    // records XML style proof - for problems in a sequence
    private void recordInFile(TableauFacade tf, String problemName, int size,
            int base, ProblemType type, Proof proof) {

        // generates the name of the file
        String nomeCompletoArquivo = tf.getProblemsOutputDir() + problemName
                + TableauFacade.FS
                + tf.generateFullProblemName(problemName, size, base, type)
                + tf.getProofTermination();

        recordInFile_Aux(tf, nomeCompletoArquivo, proof);

    }

    // records XML style proof - for isolated problems
    private void recordInFile(TableauFacade tf, String problemName, Proof proof) {

        // generates the name of the file
        String nomeCompletoArquivo = tf.getProblemsOutputDir() + problemName
                + tf.getProofTermination();

        recordInFile_Aux(tf, nomeCompletoArquivo, proof);

    }

    // records XML style proof - auxiliary method
    private void recordInFile_Aux(TableauFacade tf, String nomeCompletoArquivo,
            Proof proof) {

        // opens the file
        FileWriter f = null;
        try {
            f = new FileWriter(nomeCompletoArquivo);
        } catch (IOException e) {
            System.err.println("Problem trying to write " + nomeCompletoArquivo
                    + ".");
            e.printStackTrace();
            System.exit(0);
        }

        // generates XML text
        Document doc = new Document();

        Map m = new HashMap();
        m.put("type", "text/xsl");
        m.put("href", tf.getProblemsOutputMainDir() + TableauFacade.FS
                + tf.getXslFileName());
        doc.addContent(new ProcessingInstruction("xml-stylesheet", m));

        doc.setRootElement(proof.asXMLElement());

        // writes XML text in file
        try {
            XMLOutputter outputter = new XMLOutputter();
            outputter.setFormat(Format.getPrettyFormat());
            outputter.output(doc, f);
        } catch (Exception e) {
            System.err.println(e);
            e.printStackTrace();
            System.exit(0);
        }

        // closes file
        try {
            f.close();
        } catch (IOException e) {
            System.err.println("Problem trying to write " + nomeCompletoArquivo
                    + ".");
            e.printStackTrace();
            System.exit(0);
        }

    }

    /**
     * Recording SIMPLE PROOFS
     * 
     */

    // records simple (plain text) proof - for problems in a sequence
    private void recordSimpleProofInFile(TableauFacade tf, String problemName,
            int size, int base, ProblemType type, Proof proof) {
        String nomeCompletoArquivo = tf.getProblemsOutputDir() + problemName
                + TableauFacade.FS
                + tf.generateFullProblemName(problemName, size, base, type)
                + ".simpleproof";

        recordSimpleProofInFile_Aux(tf, nomeCompletoArquivo, proof);

    }

    // records simple (plain text) proof - for isolated problems
    private void recordSimpleProofInFile(TableauFacade tf, String problemName,
            Proof proof) {
        String nomeCompletoArquivo = tf.getProblemsOutputDir() + problemName
                + ".simpleproof";

        recordSimpleProofInFile_Aux(tf, nomeCompletoArquivo, proof);

    }

    // records simple (plain text) proof - auxiliary method
    private void recordSimpleProofInFile_Aux(TableauFacade tf,
            String completeFileName, Proof proof) {

        FileWriter f = null;
        try {
            f = new FileWriter(completeFileName);
            f.write(proof.getProofTree().toStringShort());
            f.close();
        } catch (IOException e) {
            System.err.println("Problem trying to write " + completeFileName
                    + ".");
            e.printStackTrace();
            System.exit(0);
        }
    }

    /** Breaks long formulas */

    // gets current Tableau Facade object
    TableauFacade currentTf = null;;

    pointcut TableauFacadeNew(TableauFacade tf):
        target(tf) &&
        execution(TableauFacade+.new(..));

    after(TableauFacade tf):
        TableauFacadeNew(tf){
        currentTf = tf;
    }

    pointcut simpleProof():
        cflow(target(ProofTree) &&
        call(* toString*()));

    pointcut completeProof():
         cflow(call(* Proof.asXMLElement()));

    pointcut signedFormulatoString():
        (simpleProof() || completeProof()) && 
        target (SignedFormula) &&
        call(* toString());

    String around(): signedFormulatoString(){
        String result = "";
        String s = proceed();

        if (currentTf == null) {
            System.err
                    .println("TableauFacadeProfilerAspect - Error: Tableau Facade not assigned");
            System.exit(0);
        }
        if (currentTf.getMaxNumberOfCharactersByLine() > 0) {
            for (int i = 0; i < s.length(); i++) {
                result += s.charAt(i);
                if (i > 0
                        && i % currentTf.getMaxNumberOfCharactersByLine() == 0
                        && (i != s.length() - 1)) {
                    result += System.getProperty("line.separator");
                }
            }
        }

        return result;
    }

}