/*
 * Created on 08/03/2005
 *
 */
package main.facade;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;

import logic.problem.Problem;
import logic.problem.ProblemType;
import logic.signedFormulas.SignedFormulaCreator;
import logicalSystems.classicalLogic.ClassicalRuleStructures;
import logicalSystems.classicalLogic.ClassicalSignatureFactory;
import main.strategy.Strategy;
import main.strategy.simple.SimpleStrategy;
import main.tableau.Method;
import main.tableau.Proof;
import main.tableau.Prover;

import org.jargp.ArgumentProcessor;
import org.jargp.BoolDef;
import org.jargp.IntDef;
import org.jargp.ParameterDef;
import org.jargp.StringDef;
import org.jargp.StringTracker;

import rules.structures.RulesStructure;
import util.StringUtil;

/**
 * A facade for the Tableau Prover. It configures a prover and allows us to run
 * it on some problems.
 * 
 * Its main method is configured by command-line arguments and properties from a
 * file.
 * 
 * @author Adolfo Gustavo Serra Seca Neto
 *  
 */
public class TableauFacade {

	/** Command line parameter definitions. */
	private static final ParameterDef[] PARM_DEFS = {
			new StringDef('d', "m_dSHomeDir",
			"home dir of the Directory Structure"),
			new StringDef('p', "m_propertiesFile",
			"complete name of Properties file"),
			new StringDef('r', "m_runType",
					"type of the run: sequence or instance"),
			new StringDef('n', "m_problemName",
					"name of family (for sequences) or instance of problem(s)"),
			new IntDef('b', "m_beginInstance",
					"begin instance number of a sequence"),
			new IntDef('e', "m_endInstance",
					"end instance number of a sequence"),
			new IntDef('s', "m_step", "step of a sequence"),
			new IntDef('m', "m_maxInstance",
					"max instance number of a sequence"),
			new StringDef(
					't',
					"m_problemFormat",
					"format of problem. It can be \"normal\", "
							+ "\"clausal\", \"normal_with_literals\" or \"clausal_with_literals\""),
			new BoolDef('a', "m_analyseOnly",
					"only analyse sequence or problem instance"),
			new BoolDef('o', "m_nameOnly",
					"only use name. That is, do not use numbers to generate a name."),
			new BoolDef('u', "m_showSummary",
					"shows a summary line for every problem proved or analysed."),
			new BoolDef('h', "m_saveProof", "saves proofs in files."),
			new BoolDef('l', "m_saveSimpleProof",
					"saves simplified versions of proofs in files."),
			new BoolDef('f', "m_saveInfoExecution",
					"saves information about execution in (complete) proof files."),
			new BoolDef('w', "m_onePhaseParsing",
					"sets one phase parsing for the signed formula creator."),
			new BoolDef('y', "m_saveSummary",
					"saves a summary of each proof in a summaries file."),
			new IntDef('i', "m_numberOfTimes",
					"number of times each problem is proven (default is 1)"),
			new IntDef('x', "m_maxCharactersByLine",
					"maximum number of consecutive characters in the same line"
							+ " for each signed formula (default is 80).")

	};

	// constants
	public static final String FS = System.getProperty("file.separator");

	public static final String LS = System.getProperty("line.separator");

	private static final String SEQUENCE = new String("sequence");

	private static final String INSTANCE = new String("instance");

	/** command-line attributes and default values for some of them */
	private String m_propertiesFile = "TP_properties.txt";

	private String m_dSHomeDir = "/home/kurumin/MSTP";

	private String m_runType = "sequence";

	private String m_problemName;

	private int m_beginInstance = 1;

	private int m_endInstance;

	private int m_step = 1;

	private int m_maxInstance;

	private String m_problemFormat;

	private boolean m_analyseOnly = false;

	private boolean m_nameOnly = false;

	private boolean m_helpFlag = false;

	private boolean m_showSummary = false, m_saveProof = false,
			m_saveSimpleProof = false, m_saveInfoExecution = false,
			m_onePhaseParsing = false, m_saveSummary = false;

	private int m_numberOfTimes = 1, m_maxCharactersByLine = 80;

	
	// TableauFacade object attributes

	private Prover _prover;

	private Method _method;

	private StrategyFactory _strategyFactory;

	private DirectoryStructure _ds;

	// TODO options: NOT NECESSARY?

	private String _analyserName, _problemsInputDir, _problemsOutputDir,
			_problemsOutputMainDir, _problemsTermination, _proofTermination,
			_texSeparator, _texFinalizer, _texFileTermination, _xslFileName,
			_sfCreatorLibDir, _strategyName, _classOfProblems;

	/**
	 * Creates a new tableau facade.
	 * 
	 * @param p -
	 *            the prover to be used by the facade.
	 */
	public TableauFacade(Prover p) {
		_prover = p;

		setDefaultValues();

	}

	// TODO THIS IS NOT NECESSARY ANYMORE!
	/**
	 * sets default values for attributes.
	 */
	private void setDefaultValues() {
		_analyserName = "sats4";
		_problemsOutputDir = System.getProperty("user.dir")
				+ "/../TPExe/problems/output/wagner/";
		_problemsTermination = ".prove";
		_proofTermination = ".proof";
		_xslFileName = "proof.xsl";
		_texSeparator = " & ";
		_texFinalizer = " \\\\ \\hline ";
		_texFileTermination = ".tex";
		_sfCreatorLibDir = System.getProperty("user.dir") + "/../TPExe/lib/";
		_strategyName = "SimpleStrategy";
		_strategyFactory = new StrategyFactory();
	}

	public static final String PRINT_ARGUMENTS = "print.arguments";

	public static final String TRUE = "TRUE";

	public static void main(String[] args) throws Exception {

		ConfigurationConcentrator cc = ConfigurationConcentrator.getInstance();

		// if configurated to print argumnents 
		if (cc.getProperty(PRINT_ARGUMENTS).toUpperCase().equals(TRUE)) {
			// prints arguments received
			System.err.println("Tableau Prover Facade");
			System.err.println("Arguments: " + Arrays.asList(args));
		}

		// creates argument list
		ArgumentProcessor proc = new ArgumentProcessor(PARM_DEFS);

		// creates instance of TableauFacade
		TableauFacade inst = configureTableauFacade();


		// if no arguments are supplied, assume help is needed
		if (args.length > 0) {

			// process arguments directly to instance
			List extraArgs = new ArrayList();
			proc.processArgs(args, inst);
			StringTracker xargs = proc.getArgs();
			while (xargs.hasNext()) {
				extraArgs.add(xargs.next());
			}

			inst.setDirectoryStructure(new DirectoryStructure(inst.m_dSHomeDir));
			// gets properties from Properties file
			Properties prop = getProperties(inst.getDirectoryStructure()
					.getCompletePath(DirectoryStructure.CONFIG)
					+ inst.m_propertiesFile);

			// runs tableau facade according to arguments and properties
			exec(inst, extraArgs, prop);

		} else {
			inst.m_helpFlag = true;
		}
		

		// print usage information if problem with parameters
		if (inst.m_helpFlag) {
			System.out
					.println("\nUsage: java facade.TableauFacade [-options] [extraArgs]\n"
							+ "Options are:");
			proc.listParameters(80, System.out);
		}
	}

	public static Proof runMain(String[] args) throws Exception {

		// prints arguments received
		System.err.println("Tableau Prover Facade");
		System.err.println("Arguments: " + Arrays.asList(args));

		// creates argument list
		ArgumentProcessor proc = new ArgumentProcessor(PARM_DEFS);

		// creates instance of TableauFacade
		TableauFacade inst = configureTableauFacade();

		// if no arguments are supplied, assume help is needed
		if (args.length > 0) {

			// process arguments directly to instance
			List extraArgs = new ArrayList();
			proc.processArgs(args, inst);
			StringTracker xargs = proc.getArgs();
			while (xargs.hasNext()) {
				extraArgs.add(xargs.next());
			}

			// gets properties from Properties file
			Properties prop = getProperties(inst.getDirectoryStructure()
					.getCompletePath(DirectoryStructure.CONFIG)
					+ inst.m_propertiesFile);

			// runs tableau facade according to arguments and properties
			return execProof(inst, extraArgs, prop);

		} else {
			inst.m_helpFlag = true;
		}

		// print usage information if problem with parameters
		if (inst.m_helpFlag) {
			System.out
					.println("\nUsage: java facade.TableauFacade [-options] [extraArgs]\n"
							+ "Options are:");
			proc.listParameters(80, System.out);
		}

		return null;
	}

	/**
	 * @param inst
	 * @param extraArgs
	 * @param prop
	 */
	private static void exec(TableauFacade tf, List extraArgs, Properties prop) {

		setFields(tf, prop);

		if (tf.m_runType.equals(SEQUENCE)) {

			if (tf.m_analyseOnly) {
				tf.analyseSequence(tf.m_problemName, tf.m_beginInstance,
						tf.m_endInstance, tf.m_step, tf.m_maxInstance,
						ProblemType.getInstanceByName(tf.m_problemFormat));

			} else {

				tf.proveSequence(tf.m_problemName, tf.m_beginInstance,
						tf.m_endInstance, tf.m_step, tf.m_maxInstance,
						ProblemType.getInstanceByName(tf.m_problemFormat));
			}
		} else if (tf.m_runType.equals(INSTANCE)) {

			if (tf.m_analyseOnly) {
				tf.analyseProblem(tf.m_problemName);
			} else {
				tf.prove(tf.m_problemName);
			}

		}
	}

	private static Proof execProof(TableauFacade tf, List extraArgs,
			Properties prop) {

		setFields(tf, prop);

		if (tf.m_runType.equals(SEQUENCE)) {

			if (tf.m_analyseOnly) {
				tf.analyseSequence(tf.m_problemName, tf.m_beginInstance,
						tf.m_endInstance, tf.m_step, tf.m_maxInstance,
						ProblemType.getInstanceByName(tf.m_problemFormat));

			} else {

				tf.proveSequence(tf.m_problemName, tf.m_beginInstance,
						tf.m_endInstance, tf.m_step, tf.m_maxInstance,
						ProblemType.getInstanceByName(tf.m_problemFormat));
			}
		} else if (tf.m_runType.equals(INSTANCE)) {

			if (tf.m_analyseOnly) {
				tf.analyseProblem(tf.m_problemName);
			} else {
				return tf.prove(tf.m_problemName);
			}

		}

		return null;

	}

	private static Properties getProperties(String completeFilename)
			throws Exception {

		Properties prop = new Properties();

		try {
			FileInputStream fis = new FileInputStream(completeFilename);
			prop.load(fis);
			return prop;
		} catch (Exception e) {
			System.err.println("Problems trying to read from file "
					+ completeFilename);
			System.err.println("Aborting...");
			System.exit(0);
		}

		return null;
	}

	private static TableauFacade configureTableauFacade() {
		// TODO OPTION OF SET OF RULES
		RulesStructure rules = new ClassicalRuleStructures(
				ClassicalSignatureFactory.getInstance().getNormalBXSignature())
				.getRuleStructure();
		//      RulesStructure rules =
		//      ClassicalRuleStructures.KEKESubstitionRulesWithBiimplicationAndXor
		//      .getRuleStructureOld();
		//      RulesStructure rules =
		//      ClassicalRuleStructures.KEKESubstitionRulesWithBiimplicationAndXor
		//      .getRuleStructure();
		Method m = new Method(rules);

		// This strategy is default. It can be changed by the arguments.
		Strategy s = new SimpleStrategy(m);
		Prover prover = new Prover();
		prover.setMethod(m);
		prover.setStrategy(s);

		TableauFacade tf = new TableauFacade(prover);

		tf.setMethod(m);

		return tf;
	}

	/**
	 * @param name
	 * @return
	 */

	private static void setFields(TableauFacade tf, Properties prop) {
		// new properties
		tf.p_homeDir = tf.getDirectoryStructure().getHomeDir();
		tf.p_problemsOutputDir = tf.getDirectoryStructure().getCompletePath(
				DirectoryStructure.RESULTS);
		tf.p_outputProofsDir = tf.getDirectoryStructure().getCompletePath(
				DirectoryStructure.RESULTS, DirectoryStructure.PROOFS);
		tf.p_outputSummariesDir = tf.getDirectoryStructure().getCompletePath(
				DirectoryStructure.RESULTS, DirectoryStructure.SUMMARIES);
		tf.p_classOfProblems = prop.getProperty("classOfProblems");

		// old properties

		if (prop.getProperty("analyserName") != null) {
			tf.setAnalyserName(prop.getProperty("analyserName"));
		}

		if (prop.getProperty("classOfProblems") != null) {
			tf._classOfProblems = prop.getProperty("classOfProblems");
			tf.setProblemsInputDirectory(tf.getDirectoryStructure()
					.getCompletePath(DirectoryStructure.PROBLEMS,
							prop.getProperty("typeOfProblems"),
							prop.getProperty("classOfProblems")));
		}

		if (prop.getProperty("outputProofsDir") != null) {
			tf.setProblemsOutputDirectory(tf.getDirectoryStructure()
					.getCompletePath(DirectoryStructure.RESULTS,
							prop.getProperty("outputProofsDir"),
							prop.getProperty("classOfProblems")));
			tf.setProblemsOutputMainDirectory(tf.getDirectoryStructure()
					.getCompletePath(DirectoryStructure.RESULTS));
		}

		if (prop.getProperty("XSLFileName") != null) {
			tf.setXSLFileName(prop.getProperty("XSLFileName"));
		}

		if (prop.getProperty("SFCreatorLibDir") != null) {

			tf.setSFCreatorLibDir(tf.getDirectoryStructure().getCompletePath(
					DirectoryStructure.LIB, DirectoryStructure.GENERATED));
		}

		if (prop.getProperty("proofTermination") != null) {
			tf.setProofTermination(prop.getProperty("proofTermination"));
		}

		if (prop.getProperty("strategy") != null) {
			tf.setStrategyName(prop.getProperty("strategy"));
		}

		tf.setStrategy();
	}

	private Strategy createStrategy(String strategyName, Method method) {
		return _strategyFactory.createStrategy(strategyName, method);
	}

	// prove and analyse methods

	private void proveSequence(String name, int begin, int end, int step,
			int base, ProblemType type) {
		if (base <= -1)
			base = end;

		for (int i = begin; i <= end; i += step) {
			try {
				prove(name, i, base, type);
			} catch (OutOfMemoryError e) {
				System.err.println("The prover was not able to prove "
						+ generateFullProblemName(name, i, base, type)
						+ " due to lack of memory. Exiting...");
				System.exit(0);
			}

		}

	}

	private void analyseSequence(String name, int begin, int end, int step,
			int base, ProblemType type) {
		if (base <= -1)
			base = end;

		for (int i = begin; i <= end; i += step) {
			analyseProblem(name, i, base, type);
		}
	}

	private String analyseProblem(String problemName, int size, int base,
			ProblemType type) {

		SignedFormulaCreator sfc = new SignedFormulaCreator(_analyserName);
		sfc.setLibDir(_sfCreatorLibDir);
		Problem p = configureProblem(problemName, size, base, type, sfc);

		if (m_showSummary) {
			System.out.println(generateFullProblemName(problemName, size, base,
					type)
					+ " analysed.");
		}

		return generateFamilyProblemSummary(problemName, size, base, type, p);
	}

	public String analyseProblem(String problemName) {

		SignedFormulaCreator sfc = new SignedFormulaCreator(_analyserName,
				_sfCreatorLibDir, !m_onePhaseParsing);

		Problem p = configureProblem(problemName, sfc);

		if (m_showSummary) {
			System.out.println(generateFullProblemName(problemName)
					+ " analysed.");
		}

		return generateIsolatedProblemSummary(problemName, p);
	}

	// TODO GO TO ASPECT?
	private String generateFamilyProblemSummary(String problemName, int size,
			int base, ProblemType type, Problem problem) {
		String result = "";
		result += size + _texSeparator;
		File f = new File(generateFileName(problemName, size, base, type));

		result = generateProblemSummaryPart(f, problem, result);

		return result + _texFinalizer + LS;
	}

	private String generateProblemSummaryPart(File f, Problem problem,
			String result) {
		// length in bytes
		// number of atomic formulas
		// number of composite formulas
		// number of connectives
		// complexity of formula factory
		// size of signed formula factory
		// complexity of signed formula factory = length of problem

		result += f.length() + _texSeparator;
		result += problem.getFormulaFactory().getAtomicFormulasSize()
				+ _texSeparator;
		result += problem.getFormulaFactory().getCompositeFormulasSize()
				+ _texSeparator;
		result += problem.getFormulaFactory().getConnectivesSize()
				+ _texSeparator;
		result += problem.getFormulaFactory().getComplexity() + _texSeparator;
		result += problem.getSignedFormulaFactory().getSize() + _texSeparator;
		result += problem.getSignedFormulaFactory().getComplexity();
		return result;
	}

	private String generateIsolatedProblemSummary(String problemName,
			Problem problem) {
		String result = "length (in bytes) of file" + _texSeparator
				+ "number of atomic formulas" + _texSeparator
				+ "number of composite formulas" + _texSeparator
				+ "number of connectives" + _texSeparator
				+ "complexity of formula factory" + _texSeparator
				+ "number of signed formulas" + _texSeparator
				+ "complexity of signed formula factory/lenngth of problem"
				+ _texFinalizer + LS;
		File f = new File(generateFileName2(problemName));

		result = generateProblemSummaryPart(f, problem, result);

		return result + _texFinalizer + LS;
	}

	// proves an isolated problem
	public Proof prove(String problemName) {
		return new Proof(_prover, configureProblem(problemName,
				new SignedFormulaCreator(_analyserName, _sfCreatorLibDir,
						!m_onePhaseParsing)));
	}

	// proves a family problem
	public Proof prove(String problemName, int size, int base, ProblemType type) {
		return new Proof(_prover, configureProblem(problemName, size, base,
				type, new SignedFormulaCreator(_analyserName, _sfCreatorLibDir,
						!m_onePhaseParsing)));
	}

	private Problem configureProblem(String nomeProblema, int size, int base,
			ProblemType type, SignedFormulaCreator sfc) {
		Problem p = sfc.parseFile(generateFileName(nomeProblema, size, base,
				type));
		p.setName(nomeProblema);
		p.setSize(size);
		p.setType(type);

		//        SignedFormulaList sfl = new SignedFormulaList();
		//        SignedFormula sf = p.getSignedFormulaFactory().createSignedFormula(
		//                ClassicalSigns.FALSE, SignedToUnsignedConverter.convert(p
		//                        .getFormulaFactory(), p.getFormulas()));
		//        sfl.add(sf);
		//        p.setSignedFormulaList(sfl);
		//        System.out.println(sfl);

		return p;
	}

	private Problem configureProblem(String nomeProblema,
			SignedFormulaCreator sfc) {
		Problem p = sfc
				.parseFile(generateProblemInstanceFileName(nomeProblema));
		p.setName(nomeProblema);
		return p;
	}

	private String generateProblemFileName(String nomeProblema) {
		return _problemsInputDir + nomeProblema + FS + nomeProblema;
	}

	private String generateProblemInstanceFileName(String name) {
		return _problemsInputDir + name;
	}

	public String generateFileName(String nomeProblema, int size, int base,
			ProblemType type) {
		return _problemsInputDir + nomeProblema + FS
				+ generateFullProblemName(nomeProblema, size, base, type)
				+ _problemsTermination;
	}

	public String generateFullProblemName(String nomeProblema) {
		return nomeProblema;
	}

	public String generateFileName(String nomeProblema) {
		return _problemsInputDir + FS + generateFullProblemName(nomeProblema)
				+ _problemsTermination;
	}

	public String generateFileName2(String nomeProblema) {
		return _problemsInputDir + FS + nomeProblema;
	}

	public String generateFullProblemName(String nomeProblema, int size,
			int base, ProblemType type) {
		return nomeProblema + type.toString() + "_"
				+ StringUtil.zeroes(base, size) + size;
	}

	/**
	 * @param p
	 * @return
	 */

	// attributes based on mandatory properties
	private String p_homeDir;

	private String p_problemsOutputDir;

	private String p_outputProofsDir;

	private String p_outputSummariesDir;

	private String p_classOfProblems;

	// attributes based on optional properties
	private boolean op_showProof = false;

	private String op_resultsTermination = "_results";

	private String op_analysisTermination = "_analysis";

	private String op_summaryTermination = ".tex";

	// getters

	public DirectoryStructure getDirectoryStructure() {
		return _ds;
	}

	/**
	 * gets the prover used by the tableau facade.
	 * 
	 * @return the prover
	 */
	public Prover getProver() {
		return _prover;
	}

	/**
	 * gets the method used by the prover.
	 * 
	 * @return the method
	 */
	public Method getMethod() {
		return _method;
	}

	public String getProofTermination() {
		return _proofTermination;
	}

	public String getStrategyName() {
		return _strategyName;
	}

	public String getXSLFileName() {
		return _xslFileName;
	}

	public String getTexSeparator() {
		return _texSeparator;
	}

	public String getTexFinalizer() {
		return _texFinalizer;
	}

	public String getProblemsOutputDir() {
		return _problemsOutputDir;
	}

	public String getProblemsOutputMainDir() {
		return _problemsOutputMainDir;
	}

	public String getXslFileName() {
		return _xslFileName;
	}

	public String getClassOfProblems() {
		return _classOfProblems;
	}

	/**
	 * @return Returns the m_saveProof.
	 */
	public boolean isSaveProof() {
		return m_saveProof;
	}

	/**
	 * @return Returns the m_saveSimpleProof.
	 */
	public boolean isSaveSimpleProof() {
		return m_saveSimpleProof;
	}

	/**
	 * @return Returns the m_showSummary.
	 */
	public boolean isShowSummary() {
		return m_showSummary;
	}

	/**
	 * @return Returns the m_saveInfoExecution.
	 */
	public boolean isSaveInfoExecution() {
		return m_saveInfoExecution;
	}

	/**
	 * @return Returns the m_problemName.
	 */
	public String getProblemName() {
		return m_problemName;
	}

	public String getSummariesDir() {
		return p_outputSummariesDir;
	}

	/**
	 * @return Returns the op_saveSummary.
	 */
	public boolean isSaveSummary() {
		return m_saveSummary;
	}

	/**
	 * @return Returns the op_showProof.
	 */
	public boolean isShowProof() {
		return op_showProof;
	}

	public String getSummaryTermination() {
		return op_summaryTermination;
	}

	public int getNumberOfTimes() {
		return m_numberOfTimes;
	}

	public int getMaxNumberOfCharactersByLine() {
		return m_maxCharactersByLine;
	}

	/**
	 * @return Returns the op_analysisTermination.
	 */
	public String getAnalysisTermination() {
		return op_analysisTermination;
	}

	/**
	 * @return Returns the op_resultsTermination.
	 */
	public String getResultsTermination() {
		return op_resultsTermination;
	}

	// setters
	private void setDirectoryStructure(DirectoryStructure ds) {
		_ds = ds;
	}

	/**
	 * sets the method used by the tableau facade.
	 * 
	 * @param m -
	 *            a method
	 */
	public void setMethod(Method m) {
		_method = m;
	}

	/**
	 * sets the name of the analyser.
	 * 
	 * @param newName -
	 *            name of the analyser
	 */
	public void setAnalyserName(String newName) {
		_analyserName = newName;
	}

	public void setProofTermination(String termination) {
		_proofTermination = termination;
	}

	public void setStrategyName(String name) {
		_strategyName = name;
	}

	public void setXSLFileName(String fileName) {
		_xslFileName = fileName;
	}

	private void setSFCreatorLibDir(String string) {
		_sfCreatorLibDir = string;
	}

	private void setStrategy() {
		Strategy s = createStrategy(getStrategyName(), getMethod());
		getProver().setStrategy(s);
	}

	public void setProblemsInputDirectory(String newPIDir) {
		_problemsInputDir = newPIDir;

	}

	public void setProblemsOutputDirectory(String newPODir) {
		_problemsOutputDir = newPODir;

	}

	public void setProblemsOutputMainDirectory(String newPODir) {
		_problemsOutputMainDir = newPODir;
	}

	public void setProblemsTermination(String string) {
		_problemsTermination = string;
	}

}