/*
 * Created on 24/09/2004
 *
 */
package main.proofTree;

import main.proofTree.iterator.ProofTreeBackwardLocalIterator;
import main.proofTree.iterator.ProofTreeGlobalIterator;
import main.proofTree.iterator.ProofTreeLocalIterator;

import org.jdom.Element;

/**
 * Class that represents proof trees that are built by a tableau method when
 * constructing a tableau proof.
 * 
 * @author Adolfo Gustavo Serra Seca Neto
 * 
 */
public abstract class ProofTree {
    // basic attributes of a proof tree
    protected ProofTree _parent, // parent proof tree (can be
            // null)
            _left, // left proof tree (can be null)
            _right, // right proof tree (can be null)
            _currentPT; // current proof tree (can be null)

    protected Node _root, _last, _current;

    protected ProofTree() {
    };

    /**
     * Creates a proof tree with an initial node.
     */
    public ProofTree(Node aNode) {
        _root = aNode;
        aNode.setBranch(this);
        _last = _root;
        _current = _root;
        _currentPT = this;
    }

    /**
     * Adds a node as the last node
     */
    public void addLast(Node aNode) {

        _last.setNext(aNode);
        aNode.setPrevious(_last);
        aNode.setBranch(this);
        // aNode.setNext(null);
        _last = aNode;
        _current = _last;
    }

    /**
     * Creates a new ProofTree with the node and puts it as the left subtree of
     * the current tree.
     * 
     * @param aNode
     * @return
     */
    public final ProofTree addLeft(Node aNode) {
        _left = makeInstance(aNode);
        setReferences(aNode, _left);
        setOtherStructures(_left, aNode);

        return _left;
    }

    /**
     * Creates a new ProofTree with the node and puts it as the right subtree of
     * the current tree.
     * 
     * @param aNode
     * @return
     */
    public final ProofTree addRight(Node aNode) {
        _right = makeInstance(aNode);
        setReferences(aNode, _right);
        setOtherStructures(_right, aNode);

        return _right;
    }

    /**
     * @param aNode
     * @param right
     */
    private final void setReferences(Node aNode, ProofTree newTree) {
        aNode.setBranch(newTree);
        newTree.setParent(this);
        _currentPT = newTree;
    }

    /**
     * @param aNode
     * @return an instance of a subclass of ProofTree
     */
    protected abstract ProofTree makeInstance(Node aNode);

    /**
     * sets other strcutures of a subclass of ProofTree
     * 
     * @param pt
     * @param aNode
     */
    protected abstract void setOtherStructures(ProofTree pt, Node aNode);

    /**
     * Returns the root of the tree.
     */
    public Node getRoot() {
        return _root;
    }

    /**
     * Puts pt as parent of this proof tree.
     */
    protected void setParent(ProofTree pt) {
        _parent = pt;
    }

    /**
     * @return the parent of this proof tree or null.
     */
    public ProofTree getParent() {
        return _parent;
    }

    /**
     * @return the left subtree of this proof tree or null.
     */
    public ProofTree getLeft() {
        return _left;
    }

    /**
     * @return the right subtree of this proof tree or null.
     */
    public ProofTree getRight() {
        return _right;
    }

    /**
     * @return the last node added.
     */
    public Node getLast() {
        return _last;
    }

    /**
     * @param _currentpt
     *                   The _currentPT to set.
     */
    public void setCurrentProofTree(ProofTree _currentpt) {
        _currentPT = _currentpt;
    }

    /**
     * @param other
     *                   proof tree
     */
    public void setLeft(ProofTree other) {
        _left = other;
    }

    /**
     * @param other
     *                   proof tree
     */
    public void setRight(ProofTree other) {
        _right = other;
    }

    /**
     * @return current node
     */
    public Node getCurrent() {
        return _current;
    }

    /**
     * @return local iterator (no other branches)
     */
    public ProofTreeLocalIterator getLocalIterator() {
        return new ProofTreeLocalIterator(this);
    }

    /**
     * @return backward local iterator
     */
    public ProofTreeBackwardLocalIterator getBackwardLocalIterator() {
        return new ProofTreeBackwardLocalIterator(this);
    }

    /**
     * @return global iterator (other branches can be seen)
     */
    public ProofTreeGlobalIterator getGlobalIterator() {
        return new ProofTreeGlobalIterator(this);
    }
    
    public int getNumberOfNodes(){
        int result = 0;
        Node current = _root;

        do {
            result = result + 1;
            current = current._next;
        } while (current != null);

        if (_left != null) {
            result = result + _left.getNumberOfNodes()
                    + _right.getNumberOfNodes();
        }

        return result;
        
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Object#toString()
     */
    public String toString() {
        return this.branchToString("", 0);
    }

    private String branchToString(String prefix, int spaces) {

        String result = "";

        result = spaces(spaces) + result + this.getStatus();
        Node current = _root;

        do {
            result = result + spaces(spaces) + prefix
                    + (spaces > 0 ? spaces + ":" : "") + current.toString()
                    + System.getProperty("line.separator");
            current = current._next;
        } while (current != null);

        if (_left != null) {
            result = result + _left.branchToString("l", spaces + 1)
                    + _right.branchToString("r", spaces + 1);
        }

        return result;

    }

    /**
     * @return the status of the proof tree (when applicable)
     */
    protected String getStatus() {
        return "a branch\n";
    }

    /**
     * @return a short version of the string that shows the proof tree
     */
    public String toStringShort() {
        return this.branchToStringShort("", 0);
    }

    private String branchToStringShort(String prefix, int spaces) {

        String result = "";

        result = spaces(spaces) + result + this.getStatus();
        Node current = _root;

        do {
            result = result + spaces(spaces) + prefix
                    + (spaces > 0 ? spaces + ":" : "")
                    + current.getContent().toString()
                    + System.getProperty("line.separator");
            current = current._next;
        } while (current != null);

        if (_left != null) {
            result = result + _left.branchToStringShort("l", spaces + 1)
                    + _right.branchToStringShort("r", spaces + 1);
        }

        return result;

    }

    /**
     * @param i
     * @return a number of spaces
     */
    private String spaces(int i) {
        String result = "";

        for (int j = 0; j < i; j++) {
            result = result + " ";
        }
        return result;
    }

    // XML aspect
    public Element asXMLElement() {

        Element root = new Element("proofTree");

        Element branches = new Element("branches");
        branches.setText(Integer.toString(getBranches()));
        root.addContent(branches);

        Element height = new Element("height");
        height.setText(Integer.toString(getHeight()));
        root.addContent(height);

        Element nodes = new Element("nodes");
        nodes.setText(Integer.toString(getNumberOfNodes()));
        root.addContent(nodes);

        Element status = new Element("status");
        status.setText(getStatus());
        root.addContent(status);

        root.addContent(getNodesAsXMLElements());

        return root;

    }

    private Element getNodesAsXMLElements() {

        Node current = _root;

        Element tree = new Element("tree");

        do {
            tree.addContent(current.asXMLElement());
            current = current._next;
        } while (current != null);

        if (_left != null) {

            Element left = new Element("left");
            left.addContent(_left.getNodesAsXMLElements());
            tree.addContent(left);
            Element right = new Element("right");
            right.addContent(_right.getNodesAsXMLElements());
            tree.addContent(right);

        }

        return tree;
    }

}