/*
 * Created on Oct 22, 2004
 *
 */
package main.strategy.memorySaver;

import java.util.LinkedList;

import logic.formulas.CompositeFormula;
import logic.formulas.Connective;
import logic.formulas.Formula;
import logic.formulas.FormulaList;
import logic.signedFormulas.FormulaSign;
import logic.signedFormulas.SignedFormula;
import logic.signedFormulas.SignedFormulaBuilder;
import logic.signedFormulas.SignedFormulaList;
import logicalSystems.classicalLogic.ClassicalConnectives;
import logicalSystems.classicalLogic.ClassicalRules;
import logicalSystems.classicalLogic.ClassicalSigns;
import main.proofTree.Node;
import main.proofTree.ProofTree;
import main.proofTree.SignedFormulaNode;
import main.proofTree.SignedFormulaNodeState;
import main.proofTree.iterator.ProofTreeGlobalIterator;
import main.tableau.Method;
import rules.KERuleRole;
import rules.NullRule;
import rules.Rule;
import rules.TwoPremisesOneConclusionRule;
import rules.structures.ConnectiveRoleSignRuleList;
import rules.structures.OnePremiseRuleList;
import rules.structures.PBRuleList;
import rules.structures.TopBottomRoleRuleList;
import exceptions.TPException;

/**
 * Class that represents a strategy for KE Tableaus that tries to use the
 * minimum amount of memory possible.
 * 
 * @author Adolfo Gustavo Serra Seca Neto
 *  
 */
public class MemorySaverStrategy extends AbstractMemorySaverStrategy {

	/**
	 * Creates a simple strategy for a proof method
	 */
	public MemorySaverStrategy(Method m) {
		super(m);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see strategy.Strategy#createPTInstance(proofTree.Node)
	 */
	public ProofTree createPTInstance(Node aNode) {
		return new OptimizedClassicalProofTree((SignedFormulaNode) aNode);
	}

	/**
	 * Applies as much linear rules as possible in the current branch.
	 * 
	 * @param sfb -
	 *            the formula builder
	 * @return true if has applied any rule
	 */
	protected boolean applyLinearRules(SignedFormulaBuilder sfb) {
		boolean hasApplied;
		do {
			hasApplied = applyOnePremiseRules(getThisCurrent(), sfb);
			if (getCurrent().isClosed()) {
				break;
			}

			hasApplied = applyTOPandBOTTOMrules(getThisCurrent(), sfb) || hasApplied;
			if (getCurrent().isClosed()) {
				break;
			}

			// first faster
			hasApplied = applyTwoPremiseRulesWithMap(
			// hasApplied = applyTwoPremiseRulesWithoutMapCurrentDownThenUp(
					// hasApplied = applyTwoPremiseRulesWithoutMapBottomUp(
					getThisCurrent(), sfb) || hasApplied;
			if (getCurrent().isClosed()) {
				break;
			}

		} while (hasApplied);
		return hasApplied;
	}

	private boolean applyOnePremiseRules(OptimizedClassicalProofTree proofTree,
			SignedFormulaBuilder sfb) {
		boolean hasApplied = false;

		int i = 0;

		// notice that I am using a non-recommended "i--"

		while (i < proofTree.getPBCandidates().size() && !proofTree.isClosed()) {

			// for each signed formula not used, if it accepts a one premise
			// rule,
			// apply the rule and remove it form the list of candidates.

			SignedFormula sf = (SignedFormula) proofTree.getPBCandidates().get(
					i);

			hasApplied = chooseAndApplyOnePremiseRule(proofTree, sfb, sf);
			if (hasApplied) {
				i--;
			}

			i++;
		}

		return hasApplied;
	}

	/**
	 * chooses and applies a one premise rule.
	 * 
	 * @param proofTree
	 * @param sfb
	 * @param hasApplied
	 * @param sf
	 * @return true if the rule has been applied
	 */
	private boolean chooseAndApplyOnePremiseRule(
			OptimizedClassicalProofTree proofTree, SignedFormulaBuilder sfb,
			SignedFormula sf) {
		boolean hasApplied = false;
		Rule r = chooseOnePremiseRule(proofTree, sf);

		if (r != NullRule.INSTANCE) {
			hasApplied = true;

			SignedFormulaList sfl = r.getPossibleConclusions(sfb
					.getSignedFormulaFactory(), sfb.getFormulaFactory(),
					new SignedFormulaList(sf));

			proofTree.removeFromPBCandidates(sf,
					SignedFormulaNodeState.ANALYSED);

			for (int j = 0; j < sfl.size(); j++) {
				proofTree.addLast(new SignedFormulaNode(sfl.get(j),
						SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r,
								proofTree.getNode(sf), null)));
			}
		}
		return hasApplied;
	}

	/**
	 * Applies rules with TOP or BOTTOM unary connectives.
	 * 
	 * @param proofTree
	 * @param sfb -
	 *            the formula builder
	 * @return true if some rule was applied.
	 */
	private boolean applyTOPandBOTTOMrules(
			OptimizedClassicalProofTree proofTree, SignedFormulaBuilder sfb) {

		Formula topFormula = sfb
				.createCompositeFormula(ClassicalConnectives.TOP);
		Formula bottomFormula = sfb
				.createCompositeFormula(ClassicalConnectives.BOTTOM);
		TopBottomRoleRuleList topAndBottomRules = (TopBottomRoleRuleList) getMethod()
				.getRules().get("topAndBottomRulesNew");
		// SignedFormulaList formulasWithTOPorBOTTOM = calculate(proofTree,
		// topFormula, bottomFormula);

		return chooseAndApplyTBRules(proofTree, sfb, topFormula, bottomFormula,
				topAndBottomRules
		// , formulasWithTOPorBOTTOM
		);

	}

	/**
	 * @param proofTree
	 * @param sfb
	 * @param hasApplied
	 * @param topFormula
	 * @param bottomFormula
	 * @param topAndBottomRules // *
	 * @param formulasWithTOPorBOTTOM
	 * @return
	 */
	private boolean chooseAndApplyTBRules(
			OptimizedClassicalProofTree proofTree, SignedFormulaBuilder sfb,
			Formula topFormula, Formula bottomFormula,
			TopBottomRoleRuleList topAndBottomRules
	// ,
	// SignedFormulaList formulasWithTOPorBOTTOM
	) {

		boolean hasApplied = false;

		SignedFormula sfTB;

		// while (formulasWithTOPorBOTTOM.size() > 0 && !proofTree.isClosed()) {
		while ((sfTB = ReferenceFinder.getSignedFormulaWith(proofTree
				.getPBCandidates(), topFormula, bottomFormula)) != null
				&& !proofTree.isClosed()) {
			SignedFormula aSignedFormulaWithTOPorBOTTOM = sfTB;

			// System.out.println("A sf with TOP or BOTTOM " + sfTB);
			// SignedFormula aSignedFormulaWithTOPorBOTTOM =
			// formulasWithTOPorBOTTOM
			// .get(0);

			FormulaList fl = ReferenceFinder.getSubformulaLocalReferences(
					proofTree, topFormula, aSignedFormulaWithTOPorBOTTOM);
			fl.addAll(ReferenceFinder.getSubformulaLocalReferences(proofTree,
					bottomFormula, aSignedFormulaWithTOPorBOTTOM));
			CompositeFormula cf1 = (CompositeFormula) fl.get(0);

			CompositeFormula cf1_left = null, cf1_right = null;
			if (cf1.getImmediateSubformulas().get(0) instanceof CompositeFormula) {
				cf1_left = (CompositeFormula) cf1.getImmediateSubformulas()
						.get(0);
			}
			if (cf1.getImmediateSubformulas().size() > 1
					&& cf1.getImmediateSubformulas().get(1) instanceof CompositeFormula) {
				cf1_right = (CompositeFormula) cf1.getImmediateSubformulas()
						.get(1);
			}

			Rule r = null;
			SignedFormula sf = null;

			if (cf1_left == topFormula || cf1_left == bottomFormula) {
				r = topAndBottomRules.get(cf1_left.getConnective(), cf1
						.getConnective(), KERuleRole.LEFT);
				sf = getSignedFormula(r, sfb, aSignedFormulaWithTOPorBOTTOM);
			} else {
				if (cf1_right == topFormula || cf1_right == bottomFormula) {
					r = topAndBottomRules.get(cf1_right.getConnective(), cf1
							.getConnective(), KERuleRole.RIGHT);
					sf = getSignedFormula(r, sfb, aSignedFormulaWithTOPorBOTTOM);
				}
			}

			// formulasWithTOPorBOTTOM.remove(aSignedFormulaWithTOPorBOTTOM);
			proofTree.removeFromPBCandidates(aSignedFormulaWithTOPorBOTTOM,
					SignedFormulaNodeState.ANALYSED);
			proofTree.addLast(new SignedFormulaNode(sf,
					SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r,
							proofTree.getNode(aSignedFormulaWithTOPorBOTTOM),
							null)));

			hasApplied = true;

		}
		return hasApplied;
	}

	/**
	 * gets the conclusion of a rule.
	 * 
	 * @param r -
	 *            the rule
	 * @param sfb -
	 *            the formula builder
	 * @param aFormula -
	 *            the signed formula
	 * @return the conclusion
	 */
	private SignedFormula getSignedFormula(Rule r, SignedFormulaBuilder sfb,
			SignedFormula aFormula) {
		SignedFormulaList sfl = r.getPossibleConclusions(sfb
				.getSignedFormulaFactory(), sfb.getFormulaFactory(),
				new SignedFormulaList(aFormula));
		if (sfl.size() > 0) {
			return sfl.get(0);
		} else
			throw new TPException("Rule applied but no conclusion: " + r);
	}

	/**
	 * Applies as much two premise rules as possible.
	 * 
	 * @param proofTree -
	 *            the current proof tree
	 * @param sfb -
	 *            the formula builder
	 * @return true if any rule has been applied
	 */
	private boolean applyTwoPremiseRulesWithMap(
			OptimizedClassicalProofTree proofTree, SignedFormulaBuilder sfb) {

		boolean hasApplied = false;
		// RuleList twoPremiseRules = getMethod()._rules
		// .get("twoPremiseRulesNew");

		ConnectiveRoleSignRuleList twoPremiseRulesII = (ConnectiveRoleSignRuleList) getMethod()
				.getRules().get("twoPremiseRulesNewII");

		Formula topFormula = sfb
				.createCompositeFormula(ClassicalConnectives.TOP);
		Formula bottomFormula = sfb
				.createCompositeFormula(ClassicalConnectives.BOTTOM);

		SignedFormula T_TOP = sfb.createSignedFormula(ClassicalSigns.TRUE,
				topFormula);
		SignedFormula F_BOT = sfb.createSignedFormula(ClassicalSigns.FALSE,
				bottomFormula);

		ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
				.getGlobalIterator(), firstInTheBranch = (ProofTreeGlobalIterator) proofTree
				.getGlobalIterator();

		while (!proofTree.isClosed()) {

			SignedFormulaNode sfn = null;
			if (it.hasNextLeft()) {
				sfn = (SignedFormulaNode) it.nextRight();
			} else if (firstInTheBranch.hasPrevious()) {
				sfn = (SignedFormulaNode) firstInTheBranch.previous();
			} else
				break;

			// if it is T_TOP or F_BOT, forget

			if (((SignedFormula) sfn.getContent() == T_TOP)
					|| ((SignedFormula) sfn.getContent() == F_BOT)) {
				continue;
			}

			SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
			// SignedFormulaList mainCandidates = intersect(proofTree
			// .getPBCandidates(),
			// ReferenceFinder.getParentReferences(proofTree,
			// auxCandidate.getFormula()));
			// mainCandidates.remove(auxCandidate);

			// for (int candidate = 0; candidate < mainCandidates.size();
			// candidate++) {
			SignedFormula mainCandidate;

			int begin = 0;

			while ((mainCandidate = ReferenceFinder.getSignedFormulaWith(
					proofTree.getPBCandidates(), auxCandidate.getFormula(),
					begin)) != null

			) {
				begin++;
				// System.out.println ("main candidate " + mainCandidate);
				// System.out.println ("aux candidate " + auxCandidate);
				// SignedFormula mainCandidate = mainCandidates.get(candidate);

				SignedFormulaList sflToApply = new SignedFormulaList(
						mainCandidate);
				sflToApply.add(auxCandidate);

				FormulaList fl = ReferenceFinder.getSubformulaParentReferences(
						proofTree, auxCandidate.getFormula(), mainCandidate);

				// System.out.println ("references " + fl);

				if (fl.size() >= 1) {
					CompositeFormula f = (CompositeFormula) fl.get(0);
					Connective conn = f.getConnective();

					Formula f_left = (Formula) f.getImmediateSubformulas().get(
							0);
					Formula f_right = null;
					if (f.getImmediateSubformulas().size() > 1) {
						f_right = (Formula) f.getImmediateSubformulas().get(1);
					}
					KERuleRole role = (auxCandidate.getFormula() == f_left) ? KERuleRole.LEFT
							: ((f_right != null) && (auxCandidate.getFormula() == f_right)) ? KERuleRole.RIGHT
									: null;
					FormulaSign sign = auxCandidate.getSign();

					TwoPremisesOneConclusionRule r1 = (TwoPremisesOneConclusionRule) twoPremiseRulesII
							.get(conn, role, sign);

					SignedFormulaList sflResult = r1.getPossibleConclusions(sfb
							.getSignedFormulaFactory(),
							sfb.getFormulaFactory(), sflToApply, f);
					if (sflResult != null) {
						SignedFormula conclusion = sflResult.get(0);

						SignedFormulaNode n;
						proofTree.addLast(n = new SignedFormulaNode(conclusion,
								SignedFormulaNodeState.NOT_ANALYSED,
								createOrigin(r1, proofTree
										.getNode(mainCandidate), proofTree
										.getNode(auxCandidate))));

						begin--;
						proofTree.removeFromPBCandidates(mainCandidate,
								SignedFormulaNodeState.ANALYSED);

						hasApplied = true;
					}

				}

			}
		}

		return hasApplied;
	}

	/**
	 * applies the PB rule
	 * 
	 * @param s
	 * @param sfb
	 * @return
	 */
	protected boolean applyPBRule(LinkedList s, SignedFormulaBuilder sfb) {

		boolean hasApplied = false;

		SignedFormulaList candidates = getThisCurrent().getPBCandidates();

		if (candidates.size() > 0) {
			return applyPBOnce(s, sfb, candidates);

		}

		return hasApplied;
	}

	/**
	 * @param s
	 * @param sfb
	 * @param PBRules
	 * @param candidates
	 * @return
	 */
	private boolean applyPBOnce(LinkedList s, SignedFormulaBuilder sfb,
			SignedFormulaList candidates) {
		PBRuleList PBRules = (PBRuleList) getMethod().getRules().get("PBRules");
		SignedFormula candidateChosen = candidates.get(0);

		// System.err.println(candidateChosen);
		TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule) PBRules
				.get(candidateChosen.getSign(),
						((CompositeFormula) candidateChosen.getFormula())
								.getConnective());

		SignedFormula aux = r.getAuxiliaryCandidates(
				sfb.getSignedFormulaFactory(), sfb.getFormulaFactory(),
				candidateChosen).get(0);

		SignedFormula auxOpposite = (aux.getSign().equals(ClassicalSigns.TRUE)) ? sfb
				.createSignedFormula(ClassicalSigns.FALSE, aux.getFormula())
				: sfb
						.createSignedFormula(ClassicalSigns.TRUE, aux
								.getFormula());

		// TODO verificar se concluso j est na ProofTree
		SignedFormulaList sfl = new SignedFormulaList(candidateChosen);
		sfl.add(aux);
		SignedFormula conclusion = r.getPossibleConclusions(
				sfb.getSignedFormulaFactory(), sfb.getFormulaFactory(), sfl)
				.get(0);
		getThisCurrent().removeFromPBCandidates(candidateChosen,
				SignedFormulaNodeState.ANALYSED);

		OptimizedClassicalProofTree right = (OptimizedClassicalProofTree) getCurrent()
				.addRight(new SignedFormulaNode(auxOpposite,
						SignedFormulaNodeState.NOT_ANALYSED, createOrigin(
								ClassicalRules.PB, getThisCurrent()
										.getNode(candidateChosen), null)));

		Rule complementary = PBRules.getComplementary(r);
		if (complementary != null) {
			SignedFormulaList sflComplementary = new SignedFormulaList(
					candidateChosen);
			sflComplementary.add(auxOpposite);
			SignedFormula complementaryConclusion = complementary
					.getPossibleConclusions(sfb.getSignedFormulaFactory(),
							sfb.getFormulaFactory(), sflComplementary).get(0);

			right.addLast(new SignedFormulaNode(complementaryConclusion,
					SignedFormulaNodeState.NOT_ANALYSED, createOrigin(
							complementary, getThisCurrent().getNode(candidateChosen),
							right.getNode(auxOpposite))));
		}

		OptimizedClassicalProofTree left = (OptimizedClassicalProofTree) getCurrent()
				.addLeft(new SignedFormulaNode(aux,
						SignedFormulaNodeState.NOT_ANALYSED, createOrigin(
								ClassicalRules.PB, getThisCurrent()
										.getNode(candidateChosen), null)));
		left.addLast(new SignedFormulaNode(conclusion,
				SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r, getThisCurrent()
						.getNode(candidateChosen), left.getNode(aux))));
		s.addFirst(right);
		setCurrent(left);

		return true;
	}

	/**
	 * @param auxCandidate
	 * @param mainCandidate
	 * @return
	 */
	private KERuleRole findRole(SignedFormula auxCandidate,
			SignedFormula mainCandidate) {

		if (KERuleRole.LEFT.getFormulas(mainCandidate.getFormula()).contains(
				auxCandidate.getFormula())) {
			return KERuleRole.LEFT;
		} else if (KERuleRole.RIGHT.getFormulas(mainCandidate.getFormula())
				.contains(auxCandidate.getFormula())) {
			return KERuleRole.RIGHT;
		}

		return null;
	}

	/**
	 * @param fcpt
	 * @param sfb
	 * @param r
	 * @param sf
	 * @param sfl_left
	 */
	private boolean applyEach(OptimizedClassicalProofTree fcpt,
			SignedFormulaBuilder sfb, TwoPremisesOneConclusionRule r,
			SignedFormula sf, SignedFormulaList sfl) {

		boolean hasApplied = false;
		SignedFormulaList toApply;

		for (int i = 0; i < sfl.size(); i++) {
			toApply = new SignedFormulaList();
			toApply.add(sf);
			toApply.add(sfl.get(i));

			SignedFormulaList conclusions = r.getPossibleConclusions(sfb
					.getSignedFormulaFactory(), sfb.getFormulaFactory(),
					toApply);
			System.out.println("conclusions: " + "  " + r + "  " + toApply
					+ ":" + conclusions);
			if (conclusions != null) {
				hasApplied = true;
				SignedFormula conclusion = conclusions.get(0);

				// coloca a nova na proof tree
				fcpt.addLast(new SignedFormulaNode(conclusion,
						SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r,
								fcpt.getNode(sf), fcpt.getNode(sfl.get(i)))));

			}

		}

		return hasApplied;

	}

	/**
	 * @param fcpt
	 * @param f_left
	 * @return
	 */
	private SignedFormulaList getOccurrences(OptimizedClassicalProofTree fcpt,
			Formula f) {
		SignedFormulaList sfl = new SignedFormulaList();
		ProofTreeGlobalIterator ptgi = (ProofTreeGlobalIterator) fcpt
				.getGlobalIterator();

		// goes to the last
		while (ptgi.hasNextLeft()) {
			ptgi.nextLeft();
		}

		// and then up
		while (ptgi.hasPrevious()) {
			SignedFormulaNode sfn = (SignedFormulaNode) ptgi.current();
			ptgi.previous();

			if (((SignedFormula) sfn.getContent()).getFormula().equals(f)) {
				sfl.add((SignedFormula) sfn.getContent());
			}
			// System.out.println("vendo " + sfn.getContent());
		}

		return sfl;

	}

	// /**
	// * @param cpt
	// * @param topFormula
	// * @param bottomFormula
	// * @return
	// */
	// private SignedFormulaList calculate(OptimizedClassicalProofTree cpt,
	// Formula topFormula, Formula bottomFormula) {
	// SignedFormulaList sfl = ReferenceFinder.getLocalReferences(cpt,
	// topFormula);
	// sfl.addAll(ReferenceFinder.getLocalReferences(cpt, bottomFormula));
	// sfl = intersect(cpt.getPBCandidates(), sfl);
	// return sfl;
	// }

	/**
	 * @param candidates
	 * @param sfl
	 * @return
	 */
	private SignedFormulaList intersect(SignedFormulaList candidates,
			SignedFormulaList sfl) {
		SignedFormulaList result = new SignedFormulaList();

		for (int i = 0; i < candidates.size(); i++) {
			if (sfl.contains(candidates.get(i))) {
				result.add(candidates.get(i));
			}
		}

		return result;

	}

	private Rule chooseOnePremiseRule(OptimizedClassicalProofTree cpt,
			SignedFormula sf) {

		// System.err.println(sf);

		OnePremiseRuleList onePremiseRules = (OnePremiseRuleList) getMethod()
				.getRules().get("onePremiseRules");

		if (sf.getFormula() instanceof CompositeFormula) {
			return onePremiseRules.get(sf.getSign(), ((CompositeFormula) sf
					.getFormula()).getConnective());
		}

		return NullRule.INSTANCE;

	}

}
