/*
 * Created on Oct 22, 2004
 *
 */
package main.strategy.oldsimple;

import java.util.LinkedList;

import logic.formulas.CompositeFormula;
import logic.formulas.Connective;
import logic.formulas.Formula;
import logic.formulas.FormulaList;
import logic.signedFormulas.FormulaSign;
import logic.signedFormulas.SignedFormula;
import logic.signedFormulas.SignedFormulaBuilder;
import logic.signedFormulas.SignedFormulaList;
import logicalSystems.classicalLogic.ClassicalConnectives;
import logicalSystems.classicalLogic.ClassicalRules;
import logicalSystems.classicalLogic.ClassicalSigns;
import main.proofTree.Node;
import main.proofTree.Origin;
import main.proofTree.ProofTree;
import main.proofTree.SignedFormulaNode;
import main.proofTree.SignedFormulaNodeState;
import main.proofTree.iterator.ProofTreeGlobalIterator;
import main.strategy.Strategy;
import main.tableau.Method;
import rules.KERuleRole;
import rules.NullRule;
import rules.Rule;
import rules.TwoPremisesOneConclusionRule;
import rules.structures.ConnectiveRoleSignRuleList;
import rules.structures.OnePremiseRuleList;
import rules.structures.PBRuleList;
import rules.structures.RuleList;
import rules.structures.TopBottomRoleRuleList;
import exceptions.TPException;

/**
 * Class that represents a simple strategy for KE Tableaus.
 * 
 * @author Adolfo Gustavo Serra Seca Neto
 *  
 */
public class OldSimpleStrategy extends Strategy{

    OldFormulaReferenceClassicalProofTree _current;

    /**
     * Creates a simple strategy for a proof method
     */
    public OldSimpleStrategy(Method m) {
        super(m);
        _current = null;
    }
    
    /* (non-Javadoc)
     * @see strategy.Strategy#createPTInstance(proofTree.Node)
     */
    public ProofTree createPTInstance(Node aNode) {
        return new OldFormulaReferenceClassicalProofTree((SignedFormulaNode) aNode);
    }
    

    /* (non-Javadoc)
     * @see strategy.Strategy#close(strategy.ClassicalProofTree, signedFormulasNew.SignedFormulaBuilder)
     */
    protected  ProofTree execute(ProofTree pt,
            SignedFormulaBuilder sfb) {
        
        OldFormulaReferenceClassicalProofTree proofTree = (OldFormulaReferenceClassicalProofTree) pt;

        // the prover has already loaded the problem into a proof tree!
        // if it is already closed, finish now!
        if (proofTree.isClosed())
            return proofTree;

        removeTOPandBOTTOMFromPBCandidates(proofTree);

        LinkedList branchStack = new LinkedList();
        branchStack.addFirst(proofTree);

        boolean hasApplied;
        while (!branchStack.isEmpty() && !proofTree.isClosed()) {

            _current = (OldFormulaReferenceClassicalProofTree) branchStack
                    .removeFirst();

            if ((_current.getParent() != null)
                    && (_current == _current.getParent().getRight())
                    && (!((OldClassicalProofTree) _current.getParent().getLeft())
                            .isClosed())) {
                break;
            }

            do {
                hasApplied = applyLinearRules(sfb);
                if (_current.isClosed()) {
                    break;
                } else {
                    hasApplied = applyPBRule(branchStack, sfb);
                }

            } while (hasApplied);
        }

        return proofTree;
    }

    /**
     * Applies as much linear rules as possible in the current branch.
     * 
     * @param sfb -
     *            the formula builder
     * @return true if has applied any rule
     */
    private boolean applyLinearRules(SignedFormulaBuilder sfb) {
        boolean hasApplied;
        do {
            hasApplied = applyOnePremiseRules(_current, sfb);
            if (_current.isClosed()) {
                break;
            }

            hasApplied = applyTOPandBOTTOMrules(_current, sfb) || hasApplied;
            if (_current.isClosed()) {
                break;
            }

            // first faster
            hasApplied = applyTwoPremiseRulesWithMap(
            //                        hasApplied = applyTwoPremiseRulesWithoutMapCurrentDownThenUp(
                    //                      hasApplied = applyTwoPremiseRulesWithoutMapBottomUp(
                    _current, sfb) || hasApplied;
            if (_current.isClosed()) {
                break;
            }

        } while (hasApplied);
        return hasApplied;
    }

    private boolean applyOnePremiseRules(
            OldFormulaReferenceClassicalProofTree proofTree,
            SignedFormulaBuilder sfb) {
        boolean hasApplied = false;

        int i = 0;

        // notice that I am using a non-recommended "i--"

        while (i < proofTree.getPBCandidates().size() && !proofTree.isClosed()) {

            // for each signed formula not used, if it accepts a one premise
            // rule,
            // apply the rule and remove it form the list of candidates.

            SignedFormula sf = (SignedFormula) proofTree.getPBCandidates().get(
                    i);

            hasApplied = chooseAndApplyOnePremiseRule(proofTree, sfb, sf);
            if (hasApplied) {
                i--;
            }

            i++;
        }

        return hasApplied;
    }

    /**
     * chooses and applies a one premise rule.
     * 
     * @param proofTree
     * @param sfb
     * @param hasApplied
     * @param sf
     * @return true if the rule has been applied
     */
    private boolean chooseAndApplyOnePremiseRule(
            OldFormulaReferenceClassicalProofTree proofTree,
            SignedFormulaBuilder sfb, SignedFormula sf) {
        boolean hasApplied = false;
        Rule r = chooseOnePremiseRule(proofTree, sf);

        if (r != NullRule.INSTANCE) {
            hasApplied = true;

            SignedFormulaList sfl = r.getPossibleConclusions(sfb
                    .getSignedFormulaFactory(), sfb.getFormulaFactory(),
                    new SignedFormulaList(sf));

            proofTree.removeFromPBCandidates(sf,
                    SignedFormulaNodeState.ANALYSED);

            for (int j = 0; j < sfl.size(); j++) {
                proofTree.addLast(new SignedFormulaNode(sfl.get(j),
                        SignedFormulaNodeState.NOT_ANALYSED, new Origin(r,
                                proofTree.getNode(sf), null)));
            }
        }
        return hasApplied;
    }

    /**
     * Applies rules with TOP or BOTTOM unary connectives.
     * 
     * @param proofTree
     * @param sfb -
     *            the formula builder
     * @return true if some rule was applied.
     */
    private boolean applyTOPandBOTTOMrules(
            OldFormulaReferenceClassicalProofTree proofTree,
            SignedFormulaBuilder sfb) {

        Formula topFormula = sfb
                .createCompositeFormula(ClassicalConnectives.TOP);
        Formula bottomFormula = sfb
                .createCompositeFormula(ClassicalConnectives.BOTTOM);
        TopBottomRoleRuleList topAndBottomRules = (TopBottomRoleRuleList) this._method.getRules()
                .get("topAndBottomRulesNew");
        SignedFormulaList formulasWithTOPorBOTTOM = calculate(proofTree,
                topFormula, bottomFormula);

        return chooseAndApplyTBRules(proofTree, sfb, topFormula, bottomFormula,
                topAndBottomRules, formulasWithTOPorBOTTOM);

    }

    /**
     * @param proofTree
     * @param sfb
     * @param hasApplied
     * @param topFormula
     * @param bottomFormula
     * @param topAndBottomRules
     * @param formulasWithTOPorBOTTOM
     * @return
     */
    private boolean chooseAndApplyTBRules(
            OldFormulaReferenceClassicalProofTree proofTree,
            SignedFormulaBuilder sfb, Formula topFormula,
            Formula bottomFormula, TopBottomRoleRuleList topAndBottomRules,
            SignedFormulaList formulasWithTOPorBOTTOM) {

        boolean hasApplied = false;

        while (formulasWithTOPorBOTTOM.size() > 0 && !proofTree.isClosed()) {
            SignedFormula aSignedFormulaWithTOPorBOTTOM = formulasWithTOPorBOTTOM
                    .get(0);
            FormulaList fl = proofTree.getSubformulaLocalReferences(topFormula,
                    aSignedFormulaWithTOPorBOTTOM);
            fl.addAll(proofTree.getSubformulaLocalReferences(bottomFormula,
                    aSignedFormulaWithTOPorBOTTOM));
            CompositeFormula cf1 = (CompositeFormula) fl.get(0);

            CompositeFormula cf1_left = null, cf1_right = null;
            if (cf1.getImmediateSubformulas().get(0) instanceof CompositeFormula) {
                cf1_left = (CompositeFormula) cf1.getImmediateSubformulas()
                        .get(0);
            }
            if (cf1.getImmediateSubformulas().size() > 1
                    && cf1.getImmediateSubformulas().get(1) instanceof CompositeFormula) {
                cf1_right = (CompositeFormula) cf1.getImmediateSubformulas()
                        .get(1);
            }

            Rule r = null;
            SignedFormula sf = null;

            if (cf1_left == topFormula || cf1_left == bottomFormula) {
                r = topAndBottomRules.get(cf1_left.getConnective(), cf1
                        .getConnective(), KERuleRole.LEFT);
                sf = getSignedFormula(r, sfb, aSignedFormulaWithTOPorBOTTOM);
            } else {
                if (cf1_right == topFormula || cf1_right == bottomFormula) {
                    r = topAndBottomRules.get(cf1_right.getConnective(), cf1
                            .getConnective(), KERuleRole.RIGHT);
                    sf = getSignedFormula(r, sfb, aSignedFormulaWithTOPorBOTTOM);
                }
            }

            formulasWithTOPorBOTTOM.remove(aSignedFormulaWithTOPorBOTTOM);
            proofTree.removeFromPBCandidates(aSignedFormulaWithTOPorBOTTOM,
                    SignedFormulaNodeState.ANALYSED);
            proofTree.addLast(new SignedFormulaNode(sf,
                    SignedFormulaNodeState.NOT_ANALYSED, new Origin(r,
                            proofTree.getNode(aSignedFormulaWithTOPorBOTTOM),
                            null)));

            hasApplied = true;

            if (topFormula.isStrictSubformula(sf.getFormula())
                    || bottomFormula.isStrictSubformula(sf.getFormula())) {
                if (!formulasWithTOPorBOTTOM.contains(sf)) {
                    formulasWithTOPorBOTTOM.add(0, sf);
                }
            }
        }
        return hasApplied;
    }

    /**
     * gets the conclusion of a rule.
     * 
     * @param r -
     *            the rule
     * @param sfb -
     *            the formula builder
     * @param aFormula -
     *            the signed formula
     * @return the conclusion
     */
    private SignedFormula getSignedFormula(Rule r, SignedFormulaBuilder sfb,
            SignedFormula aFormula) {
        SignedFormulaList sfl = r.getPossibleConclusions(sfb
                .getSignedFormulaFactory(), sfb.getFormulaFactory(),
                new SignedFormulaList(aFormula));
        if (sfl.size() > 0) {
            return sfl.get(0);
        } else
            throw new TPException("Rule applied but no conclusion: " + r);
    }

    /**
     * Applies as much two premise rules as possible.
     * 
     * @param proofTree -
     *            the current proof tree
     * @param sfb -
     *            the formula builder
     * @return true if any rule has been applied
     */
    private boolean applyTwoPremiseRulesWithMap(
            OldFormulaReferenceClassicalProofTree proofTree,
            SignedFormulaBuilder sfb) {

        boolean hasApplied = false;
        //        RuleList twoPremiseRules = this._method.getRules()
        //                .get("twoPremiseRulesNew");

        ConnectiveRoleSignRuleList twoPremiseRulesII = (ConnectiveRoleSignRuleList) this._method.getRules()
                .get("twoPremiseRulesNewII");

        Formula topFormula = sfb
                .createCompositeFormula(ClassicalConnectives.TOP);
        Formula bottomFormula = sfb
                .createCompositeFormula(ClassicalConnectives.BOTTOM);

        SignedFormula T_TOP = sfb.createSignedFormula(ClassicalSigns.TRUE,
                topFormula);
        SignedFormula F_BOT = sfb.createSignedFormula(ClassicalSigns.FALSE,
                bottomFormula);

        ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
                .getGlobalIterator(), firstInTheBranch = (ProofTreeGlobalIterator) proofTree
                .getGlobalIterator();

        while (!proofTree.isClosed()) {

            SignedFormulaNode sfn = null;
            if (it.hasNextLeft()) {
                sfn = (SignedFormulaNode) it.nextRight();
            } else if (firstInTheBranch.hasPrevious()) {
                sfn = (SignedFormulaNode) firstInTheBranch.previous();
            } else
                break;

            // if it is T_TOP or F_BOT, forget

            if (((SignedFormula) sfn.getContent() == T_TOP)
                    || ((SignedFormula) sfn.getContent() == F_BOT)) {
                continue;
            }

            SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
            SignedFormulaList mainCandidates = intersect(proofTree
                    .getPBCandidates(), proofTree
                    .getParentReferences(auxCandidate.getFormula()));
            mainCandidates.remove(auxCandidate);

            for (int candidate = 0; candidate < mainCandidates.size(); candidate++) {
                SignedFormula mainCandidate = mainCandidates.get(candidate);

                //                for (int i = 0; i < twoPremiseRules.size(); i++) {
                //                    TwoPremisesOneConclusionRule r =
                // (TwoPremisesOneConclusionRule) twoPremiseRules
                //                            .get(i);

                SignedFormulaList sflToApply = new SignedFormulaList(
                        mainCandidate);
                sflToApply.add(auxCandidate);

                FormulaList fl = proofTree.getSubformulaParentReferences(
                        auxCandidate.getFormula(), mainCandidate);
                if (fl.size() >= 1) {
                    //                        System.err.println (mainCandidate + " " + auxCandidate );
                    CompositeFormula f = (CompositeFormula) fl.get(0);
                    Connective conn = f.getConnective();
                    //                    System.err.println(fl + " " + conn);

                    Formula f_left = (Formula) f.getImmediateSubformulas().get(
                            0);
                    Formula f_right = null;
                    if (f.getImmediateSubformulas().size() > 1) {
                        f_right = (Formula) f.getImmediateSubformulas().get(1);
                    }
                    KERuleRole role = (auxCandidate.getFormula() == f_left) ? KERuleRole.LEFT
                            : ((f_right != null) && (auxCandidate.getFormula() == f_right)) ? KERuleRole.RIGHT
                                    : null;
                    FormulaSign sign = auxCandidate.getSign();

                    //                      System.err.println(role);

                    TwoPremisesOneConclusionRule r1 = (TwoPremisesOneConclusionRule) twoPremiseRulesII
                            .get(conn, role, sign);
                    //                      System.err.println(r1);

                    SignedFormulaList sflResult = r1.getPossibleConclusions(sfb
                            .getSignedFormulaFactory(),
                            sfb.getFormulaFactory(), sflToApply, f);
                    if (sflResult != null) {
                        SignedFormula conclusion = sflResult.get(0);

                        SignedFormulaNode n;
                        proofTree.addLast(n = new SignedFormulaNode(conclusion,
                                SignedFormulaNodeState.NOT_ANALYSED,
                                new Origin(r1,
                                        proofTree.getNode(mainCandidate),
                                        proofTree.getNode(auxCandidate))));

                        proofTree.removeFromPBCandidates(mainCandidate,
                                SignedFormulaNodeState.ANALYSED);

                        hasApplied = true;
                    }

                }

                //                }
            }
        }

        return hasApplied;
    }

    /**
     * applies the PB rule
     * 
     * @param s
     * @param sfb
     * @return
     */
    private boolean applyPBRule(LinkedList s, SignedFormulaBuilder sfb) {

        boolean hasApplied = false;

        SignedFormulaList candidates = _current.getPBCandidates();

        if (candidates.size() > 0) {
            return applyPBOnce(s, sfb, candidates);

        }

        return hasApplied;
    }

    /**
     * @param s
     * @param sfb
     * @param PBRules
     * @param candidates
     * @return
     */
    private boolean applyPBOnce(LinkedList s, SignedFormulaBuilder sfb,
            SignedFormulaList candidates) {
        PBRuleList PBRules = (PBRuleList) this._method.getRules().get("PBRules");
        SignedFormula candidateChosen = candidates.get(0);

        //                System.err.println(candidateChosen);
        TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule) PBRules
                .get(candidateChosen.getSign(),
                        ((CompositeFormula) candidateChosen.getFormula())
                                .getConnective());

        SignedFormula aux = r.getAuxiliaryCandidates(
                sfb.getSignedFormulaFactory(), sfb.getFormulaFactory(),
                candidateChosen).get(0);

        SignedFormula auxOpposite = (aux.getSign().equals(ClassicalSigns.TRUE)) ? sfb
                .createSignedFormula(ClassicalSigns.FALSE, aux.getFormula())
                : sfb
                        .createSignedFormula(ClassicalSigns.TRUE, aux
                                .getFormula());

        // TODO verificar se concluso j est na ProofTree
        SignedFormulaList sfl = new SignedFormulaList(candidateChosen);
        sfl.add(aux);
        SignedFormula conclusion = r.getPossibleConclusions(
                sfb.getSignedFormulaFactory(), sfb.getFormulaFactory(), sfl)
                .get(0);
        _current.removeFromPBCandidates(candidateChosen,
                SignedFormulaNodeState.ANALYSED);

        OldFormulaReferenceClassicalProofTree right = (OldFormulaReferenceClassicalProofTree) _current
                .addRight(new SignedFormulaNode(auxOpposite,
                        SignedFormulaNodeState.NOT_ANALYSED, new Origin(
                                ClassicalRules.PB, _current
                                        .getNode(candidateChosen), null)));

        Rule complementary = PBRules.getComplementary(r);
        if (complementary != null) {
            SignedFormulaList sflComplementary = new SignedFormulaList(
                    candidateChosen);
            sflComplementary.add(auxOpposite);
            SignedFormula complementaryConclusion = complementary
                    .getPossibleConclusions(sfb.getSignedFormulaFactory(),
                            sfb.getFormulaFactory(), sflComplementary).get(0);

            right.addLast(new SignedFormulaNode(complementaryConclusion,
                    SignedFormulaNodeState.NOT_ANALYSED, new Origin(
                            complementary, _current.getNode(candidateChosen),
                            right.getNode(auxOpposite))));
        }

        OldFormulaReferenceClassicalProofTree left = (OldFormulaReferenceClassicalProofTree) _current
                .addLeft(new SignedFormulaNode(aux,
                        SignedFormulaNodeState.NOT_ANALYSED, new Origin(
                                ClassicalRules.PB, _current
                                        .getNode(candidateChosen), null)));
        left.addLast(new SignedFormulaNode(conclusion,
                SignedFormulaNodeState.NOT_ANALYSED, new Origin(r, _current
                        .getNode(candidateChosen), left.getNode(aux))));
        s.addFirst(_current);
        s.addFirst(right);
        _current = left;

        return true;
    }

    /**
     * @param auxCandidate
     * @param mainCandidate
     * @return
     */
    private KERuleRole findRole(SignedFormula auxCandidate,
            SignedFormula mainCandidate) {

        if (KERuleRole.LEFT.getFormulas(mainCandidate.getFormula()).contains(
                auxCandidate.getFormula())) {
            return KERuleRole.LEFT;
        } else if (KERuleRole.RIGHT.getFormulas(mainCandidate.getFormula())
                .contains(auxCandidate.getFormula())) {
            return KERuleRole.RIGHT;
        }

        return null;
    }

    /**
     * @param fcpt
     * @param sfb
     * @param r
     * @param sf
     * @param sfl_left
     */
    private boolean applyEach(OldFormulaReferenceClassicalProofTree fcpt,
            SignedFormulaBuilder sfb, TwoPremisesOneConclusionRule r,
            SignedFormula sf, SignedFormulaList sfl) {

        boolean hasApplied = false;
        SignedFormulaList toApply;

        for (int i = 0; i < sfl.size(); i++) {
            toApply = new SignedFormulaList();
            toApply.add(sf);
            toApply.add(sfl.get(i));

            SignedFormulaList conclusions = r.getPossibleConclusions(sfb
                    .getSignedFormulaFactory(), sfb.getFormulaFactory(),
                    toApply);
            System.out.println("conclusions: " + "  " + r + "  " + toApply
                    + ":" + conclusions);
            if (conclusions != null) {
                hasApplied = true;
                SignedFormula conclusion = conclusions.get(0);

                // coloca a nova na proof tree
                fcpt.addLast(new SignedFormulaNode(conclusion,
                        SignedFormulaNodeState.NOT_ANALYSED, new Origin(r, fcpt
                                .getNode(sf), fcpt.getNode(sfl.get(i)))));

            }

        }

        return hasApplied;

    }

    /**
     * @param fcpt
     * @param f_left
     * @return
     */
    private SignedFormulaList getOccurrences(
            OldFormulaReferenceClassicalProofTree fcpt, Formula f) {
        SignedFormulaList sfl = new SignedFormulaList();
        ProofTreeGlobalIterator ptgi = (ProofTreeGlobalIterator) fcpt
                .getGlobalIterator();

        // goes to the last
        while (ptgi.hasNextLeft()) {
            ptgi.nextLeft();
        }

        // and then up
        while (ptgi.hasPrevious()) {
            SignedFormulaNode sfn = (SignedFormulaNode) ptgi.current();
            ptgi.previous();

            if (((SignedFormula) sfn.getContent()).getFormula().equals(f)) {
                sfl.add((SignedFormula) sfn.getContent());
            }
            //            System.out.println("vendo " + sfn.getContent());
        }

        return sfl;

    }

    /**
     * @param cpt
     * @param topFormula
     * @param bottomFormula
     * @return
     */
    private SignedFormulaList calculate(OldFormulaReferenceClassicalProofTree cpt,
            Formula topFormula, Formula bottomFormula) {
        SignedFormulaList sfl = cpt.getLocalReferences(topFormula);
        sfl.addAll(cpt.getLocalReferences(bottomFormula));
        sfl = intersect(cpt.getPBCandidates(), sfl);
        return sfl;
    }

    /**
     * @param candidates
     * @param sfl
     * @return
     */
    private SignedFormulaList intersect(SignedFormulaList candidates,
            SignedFormulaList sfl) {
        SignedFormulaList result = new SignedFormulaList();

        for (int i = 0; i < candidates.size(); i++) {
            if (sfl.contains(candidates.get(i))) {
                result.add(candidates.get(i));
            }
        }

        return result;

    }

    /**
     * @param cpt
     */
    private void removeTOPandBOTTOMFromPBCandidates(OldClassicalProofTree cpt) {
        // remove T TOP and F BOT -- we already know they are useless as main
        // premises of rules
        cpt.removeFromPBCandidates(0);
        cpt.removeFromPBCandidates(0);
    }

    private Rule chooseOnePremiseRule(OldFormulaReferenceClassicalProofTree cpt, SignedFormula sf) {

        //        System.err.println(sf);

        OnePremiseRuleList onePremiseRules = (OnePremiseRuleList) this._method.getRules()
                .get("onePremiseRules");

        if (sf.getFormula() instanceof CompositeFormula) {
            return onePremiseRules.get(sf.getSign(), ((CompositeFormula) sf
                    .getFormula()).getConnective());
        }

        return NullRule.INSTANCE;

    }

    //    /**
    //     * Applies as much one premise rules as possible
    //     *
    //     * @param proofTree -
    //     * the current proof tree
    //     * @param sfb -
    //     * the formula builder
    //     * @return true if any rule has been applied
    //     */
    //    private boolean applyOnePremiseRulesWithIterator(
    //            FormulaReferenceClassicalProofTree proofTree,
    //            SignedFormulaBuilder sfb) {
    //        /*
    //         *
    //         * iterate over current branch
    //         *
    //         * for each formula, see if a 1-premise rule is applicable. If it is,
    //         * apply it and remove it from list of formulas for which 1-premise
    //         * rules can be applied.
    //         *
    //         */
    //        boolean hasApplied = false;
    //
    //        Iterator it = proofTree.getPBCandidates().iterator();
    //
    //        try {
    //            while (it.hasNext()) {
    //                SignedFormula sf = (SignedFormula) it.next();
    //
    //                hasApplied = chooseAndApplyOnePremiseRule(proofTree, sfb, sf);
    //
    //            }
    //        } catch (ConcurrentModificationException e) {
    //            // No problem. Same thread.
    //            // TODO it would be wrong to write a program that depended on this
    //            // exception for its correctness: ConcurrentModificationException
    //            // should be used only to detect bugs.
    //        }
    //
    //        return hasApplied;
    //    }

    //    /**
    //     * Applies as much two premise rules as possible.
    //     *
    //     * @param proofTree -
    //     * the current proof tree
    //     * @param sfb -
    //     * the formula builder
    //     * @return true if any rule has been applied
    //     */
    //    private boolean applyTwoPremiseRules(
    //            FormulaReferenceClassicalProofTree proofTree,
    //            SignedFormulaBuilder sfb) {
    //
    //        boolean hasApplied = false;
    //        RuleList twoPremiseRules = this._method.getRules()
    //                .get("twoPremiseRulesNew");
    //
    //        ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
    //                .getGlobalIterator();
    //
    //// AQUI
    //        // TODO Deixar de ir de baixo para cima. Criar novo iterador.
    //        // goes to the last
    //        while (it.hasNextLeft()) {
    //            it.nextLeft();
    //        }
    //        // ProofTreeBasicIterator it = fcpt.getGlobalIterator();
    //        //TODO deixar de olhar o F BOT
    //        while (it.hasPrevious() && !proofTree.isClosed()) {
    //            SignedFormulaNode sfn = (SignedFormulaNode) it.current();
    //            it.previous();
    //
    //            // System.err.println(" FAss do n " + sfn.getContent());
    //            SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
    //            // System.err.println(" Ref  frmula na rvore "
    //            // + fcpt.getParentReferences(auxCandidate.getFormula()));
    //            SignedFormulaList cand = intersect(proofTree.getPBCandidates(),
    //                    proofTree.getParentReferences(auxCandidate.getFormula()));
    //            cand.remove(auxCandidate);
    //            // System.err.println(" Ref cand  frmula na rvore " + cand);
    //
    //            for (int candidate = 0; candidate < cand.size(); candidate++) {
    //                SignedFormula mainCandidate = cand.get(candidate);
    //
    //                // TODO no funciona porque pode ser um conectivo interior
    //                // Connective mainConnective = ((CompositeFormula) mainCandidate
    //                // .getFormula()).getConnective();
    //                // System.err.print(mainConnective + " \n");
    //
    //                for (int i = 0; i < twoPremiseRules.size(); i++) {
    //                    TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule)
    // twoPremiseRules
    //                            .get(i);
    //
    //                    // System.err.print(r + " ");
    //
    //                    SignedFormulaList sflToApply = new SignedFormulaList(
    //                            mainCandidate);
    //                    sflToApply.add(auxCandidate);
    //                    // System.err.print(sflToApply);
    //
    //                    // TODO no funciona porque pode ser subfrmula
    //                    // KERuleRole auxRole = findRole(auxCandidate,
    //                    // mainCandidate);
    //
    //                    // System.err.print(auxRole + " ");
    //
    //                    SignedFormulaList sflResult = r.getPossibleConclusions(sfb
    //                            .getSignedFormulaFactory(),
    //                            sfb.getFormulaFactory(), sflToApply);
    //                    // System.err.println(" result " + sflResult);
    //                    if (sflResult != null) {
    //                        SignedFormula conclusion = sflResult.get(0);
    //
    //                        // coloca a nova na proof tree
    //                        SignedFormulaNode n;
    //                        proofTree.addLast(n = new SignedFormulaNode(conclusion,
    //                                SignedFormulaNodeState.NOT_ANALYSED,
    //                                new Origin(r, proofTree.getNode(mainCandidate),
    //                                        proofTree.getNode(auxCandidate))));
    //
    //                        // System.err.println ("Nodeadded " +n + " at " + fcpt);
    //
    //                        proofTree.removeFromPBCandidates(mainCandidate,
    //                                SignedFormulaNodeState.ANALYSED);
    //
    //                        hasApplied = true;
    //                    }
    //
    //                }
    //
    //            }
    //
    //        }
    //
    //        return hasApplied;
    //    }

    /**
     * Applies as much two premise rules as possible.
     * 
     * @param proofTree -
     *            the current proof tree
     * @param sfb -
     *            the formula builder
     * @return true if any rule has been applied
     */
    private boolean applyTwoPremiseRulesWithoutMapBottomUp(
            OldFormulaReferenceClassicalProofTree proofTree,
            SignedFormulaBuilder sfb) {

        boolean hasApplied = false;

        RuleList twoPremiseRules = this._method.getRules()
                .get("twoPremiseRulesNew");

        ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
                .getGlobalIterator();

        Formula topFormula = sfb
                .createCompositeFormula(ClassicalConnectives.TOP);
        Formula bottomFormula = sfb
                .createCompositeFormula(ClassicalConnectives.BOTTOM);

        SignedFormula T_TOP = sfb.createSignedFormula(ClassicalSigns.TRUE,
                topFormula);
        SignedFormula F_BOT = sfb.createSignedFormula(ClassicalSigns.FALSE,
                bottomFormula);

        // AQUI
        // TODO Deixar de ir de baixo para cima. Criar novo iterador.
        // goes to the last
        while (it.hasNextLeft()) {
            it.nextLeft();
        }
        // ProofTreeBasicIterator it = fcpt.getGlobalIterator();
        //TODO deixar de olhar o F BOT
        while (it.hasPrevious() && !proofTree.isClosed()) {
            SignedFormulaNode sfn = (SignedFormulaNode) it.current();
            it.previous();

            if (((SignedFormula) sfn.getContent() == T_TOP)
                    || ((SignedFormula) sfn.getContent() == F_BOT)) {
                continue;
            }

            // System.err.println(" FAss do n " + sfn.getContent());
            SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
            // System.err.println(" Ref  frmula na rvore "
            // + fcpt.getParentReferences(auxCandidate.getFormula()));
            SignedFormulaList cand = intersect(proofTree.getPBCandidates(),
                    proofTree.getParentReferences(auxCandidate.getFormula()));
            cand.remove(auxCandidate);
            // System.err.println(" Ref cand  frmula na rvore " + cand);

            for (int candidate = 0; candidate < cand.size(); candidate++) {
                SignedFormula mainCandidate = cand.get(candidate);

                // TODO no funciona porque pode ser um conectivo interior
                // Connective mainConnective = ((CompositeFormula) mainCandidate
                // .getFormula()).getConnective();
                // System.err.print(mainConnective + " \n");

                for (int i = 0; i < twoPremiseRules.size(); i++) {
                    TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule) twoPremiseRules
                            .get(i);

                    // System.err.print(r + " ");

                    SignedFormulaList sflToApply = new SignedFormulaList(
                            mainCandidate);
                    sflToApply.add(auxCandidate);
                    // System.err.print(sflToApply);

                    // TODO no funciona porque pode ser subfrmula
                    // KERuleRole auxRole = findRole(auxCandidate,
                    // mainCandidate);

                    // System.err.print(auxRole + " ");

                    SignedFormulaList sflResult = r.getPossibleConclusions(sfb
                            .getSignedFormulaFactory(),
                            sfb.getFormulaFactory(), sflToApply);
                    // System.err.println(" result " + sflResult);
                    if (sflResult != null) {
                        SignedFormula conclusion = sflResult.get(0);

                        // coloca a nova na proof tree
                        SignedFormulaNode n;
                        proofTree.addLast(n = new SignedFormulaNode(conclusion,
                                SignedFormulaNodeState.NOT_ANALYSED,
                                new Origin(r, proofTree.getNode(mainCandidate),
                                        proofTree.getNode(auxCandidate))));

                        // System.err.println ("Nodeadded " +n + " at " + fcpt);

                        proofTree.removeFromPBCandidates(mainCandidate,
                                SignedFormulaNodeState.ANALYSED);

                        hasApplied = true;
                    }

                }

            }

        }

        return hasApplied;
    }

    /**
     * Applies as much two premise rules as possible.
     * 
     * @param proofTree -
     *            the current proof tree
     * @param sfb -
     *            the formula builder
     * @return true if any rule has been applied
     */
    private boolean applyTwoPremiseRulesWithoutMapCurrentDownThenUp(
            OldFormulaReferenceClassicalProofTree proofTree,
            SignedFormulaBuilder sfb) {

        // TODO REMOVE
        ConnectiveRoleSignRuleList twoPremiseRulesII = (ConnectiveRoleSignRuleList) this._method.getRules()
                .get("twoPremiseRulesNewII");

        boolean hasApplied = false;
        RuleList twoPremiseRules = this._method.getRules()
                .get("twoPremiseRulesNew");

        ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
                .getGlobalIterator(), firstInTheBranch = (ProofTreeGlobalIterator) proofTree
                .getGlobalIterator();

        Formula topFormula = sfb
                .createCompositeFormula(ClassicalConnectives.TOP);
        Formula bottomFormula = sfb
                .createCompositeFormula(ClassicalConnectives.BOTTOM);

        SignedFormula T_TOP = sfb.createSignedFormula(ClassicalSigns.TRUE,
                topFormula);
        SignedFormula F_BOT = sfb.createSignedFormula(ClassicalSigns.FALSE,
                bottomFormula);

        while (!proofTree.isClosed()) {

            SignedFormulaNode sfn = null;
            if (it.hasNextLeft()) {
                sfn = (SignedFormulaNode) it.nextRight();
            } else if (firstInTheBranch.hasPrevious()) {
                sfn = (SignedFormulaNode) firstInTheBranch.previous();
            } else
                break;

            if (((SignedFormula) sfn.getContent() == T_TOP)
                    || ((SignedFormula) sfn.getContent() == F_BOT)) {
                continue;
            }

            SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
            SignedFormulaList mainCandidates = intersect(proofTree
                    .getPBCandidates(), proofTree
                    .getParentReferences(auxCandidate.getFormula()));
            mainCandidates.remove(auxCandidate);

            for (int candidate = 0; candidate < mainCandidates.size(); candidate++) {
                SignedFormula mainCandidate = mainCandidates.get(candidate);

                for (int i = 0; i < twoPremiseRules.size(); i++) {
                    TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule) twoPremiseRules
                            .get(i);

                    SignedFormulaList sflToApply = new SignedFormulaList(
                            mainCandidate);
                    sflToApply.add(auxCandidate);

                    //                                        FormulaList fl = proofTree.getSubformulaParentReferences(
                    //                                                auxCandidate.getFormula(), mainCandidate);
                    //                                        if (fl.size() > 1) {
                    //                                            // System.err.println (mainCandidate + " " +
                    //                                            // auxCandidate );
                    //                                            CompositeFormula f = (CompositeFormula) fl.get(0);
                    //                                            Connective conn = f.getConnective();
                    //                                            // System.err.println(fl + " " + conn);
                    //                    
                    //                                            Formula f_left = (Formula) f.getImmediateSubformulas()
                    //                                                    .get(0);
                    //                                            Formula f_right = null;
                    //                                            if (f.getImmediateSubformulas().size() > 1) {
                    //                                                f_right = (Formula) f.getImmediateSubformulas()
                    //                                                        .get(1);
                    //                                            }
                    //                                            KERuleRole role = (auxCandidate.getFormula() == f_left) ?
                    //                     KERuleRole.LEFT
                    //                                                    : ((f_right != null) && (auxCandidate
                    //                                                            .getFormula() == f_right)) ? KERuleRole.RIGHT
                    //                                                            : null;
                    //                    
                    //                                            // System.err.println(role);
                    //                                            FormulaSign sign = auxCandidate.getSign();
                    //                    
                    //                                            TwoPremisesOneConclusionRule r1 =
                    //                     (TwoPremisesOneConclusionRule) twoPremiseRulesII
                    //                                                    .get(conn, role, sign);
                    //// System.err.println(r1);
                    //                    
                    //                                            SignedFormulaList sflResultX = r1
                    //                                                    .getPossibleConclusions(sfb
                    //                                                            .getSignedFormulaFactory(), sfb
                    //                                                            .getFormulaFactory(), sflToApply, f);
                    //
                    //                                            // repeteco - incio
                    //                                            if (sflResultX != null) {
                    //                                                SignedFormula conclusion = sflResultX.get(0);
                    //
                    //                                                SignedFormulaNode n;
                    //                                                proofTree.addLast(n = new SignedFormulaNode(conclusion,
                    //                                                        SignedFormulaNodeState.NOT_ANALYSED,
                    //                                                        new Origin(r1, proofTree.getNode(mainCandidate),
                    //                                                                proofTree.getNode(auxCandidate))));
                    //
                    //                                                proofTree.removeFromPBCandidates(mainCandidate,
                    //                                                        SignedFormulaNodeState.ANALYSED);
                    //
                    //                                                hasApplied = true;
                    //                                            }
                    //                                            // repeteco - fim
                    //                                        }

                    SignedFormulaList sflResult = r.getPossibleConclusions(sfb
                            .getSignedFormulaFactory(),
                            sfb.getFormulaFactory(), sflToApply);
                    if (sflResult != null) {
                        SignedFormula conclusion = sflResult.get(0);

                        SignedFormulaNode n;
                        proofTree.addLast(n = new SignedFormulaNode(conclusion,
                                SignedFormulaNodeState.NOT_ANALYSED,
                                new Origin(r, proofTree.getNode(mainCandidate),
                                        proofTree.getNode(auxCandidate))));

                        proofTree.removeFromPBCandidates(mainCandidate,
                                SignedFormulaNodeState.ANALYSED);

                        hasApplied = true;
                    }

                }
            }
        }

        return hasApplied;
    }


}