/*
 * Created on Oct 22, 2004
 *
 */
package main.strategy.simple;

import java.util.LinkedList;

import logic.formulas.CompositeFormula;
import logic.formulas.Connective;
import logic.formulas.Formula;
import logic.formulas.FormulaList;
import logic.signedFormulas.FormulaSign;
import logic.signedFormulas.SignedFormula;
import logic.signedFormulas.SignedFormulaBuilder;
import logic.signedFormulas.SignedFormulaList;
import logicalSystems.classicalLogic.ClassicalConnectives;
import logicalSystems.classicalLogic.ClassicalRules;
import logicalSystems.classicalLogic.ClassicalSigns;
import main.proofTree.Node;
import main.proofTree.Origin;
import main.proofTree.ProofTree;
import main.proofTree.SignedFormulaNode;
import main.proofTree.SignedFormulaNodeState;
import main.proofTree.iterator.ProofTreeGlobalIterator;
import main.strategy.ClassicalProofTree;
import main.tableau.Method;
import rules.KERuleRole;
import rules.NullRule;
import rules.Rule;
import rules.TwoPremisesOneConclusionRule;
import rules.structures.ConnectiveRoleSignRuleList;
import rules.structures.OnePremiseRuleList;
import rules.structures.PBRuleList;
import rules.structures.RuleList;
import rules.structures.TopBottomRoleRuleList;
import exceptions.TPException;

/**
 * Class that represents a simple strategy for KE Tableaus.
 * 
 * Uses FormulaReferenceClassicalProofTree objects that keep references of
 * formulas and subformulas.
 * 
 * @author Adolfo Gustavo Serra Seca Neto
 *  
 */
public class SimpleStrategy extends AbstractSimpleStrategy {

	/**
	 * Creates a simple strategy for a proof method
	 * @param m a method
	 */
	public SimpleStrategy(Method m) {
		super(m);
	}

	/* (non-Javadoc)
	 * @see main.strategy.AbstractStrategy#createOrigin(rules.Rule, main.proofTree.SignedFormulaNode, main.proofTree.SignedFormulaNode)
	 */
	protected Origin createOrigin(Rule rule, SignedFormulaNode main,
			SignedFormulaNode auxiliary) {
		return new Origin(rule, main, auxiliary);
	}

	/* (non-Javadoc)
	 * @see main.strategy.Strategy#createPTInstance(main.proofTree.Node)
	 */
	public ProofTree createPTInstance(Node aNode) {
		return new FormulaReferenceClassicalProofTree((SignedFormulaNode) aNode);
	}

	/* (non-Javadoc)
	 * @see main.strategy.AbstractStrategy#isRightBranchOfOpenLeftBranch()
	 */
	public boolean isRightBranchOfOpenLeftBranch() {
		return (getThisCurrent().getParent() != null)
				&& (getThisCurrent() == getThisCurrent().getParent().getRight())
				&& (!((FormulaReferenceClassicalProofTree) getThisCurrent()
						.getParent().getLeft()).isClosed());
	}

	/* (non-Javadoc)
	 * @see main.strategy.AbstractStrategy#maybeEraseLeft()
	 */
	protected void maybeEraseLeft() {
	}


	/* (non-Javadoc)
	 * @see main.strategy.simple.AbstractSimpleStrategy#applyOnePremiseRules(main.strategy.ClassicalProofTree, logic.signedFormulas.SignedFormulaBuilder)
	 */
	protected boolean applyOnePremiseRules(
			ClassicalProofTree proofTree,
			SignedFormulaBuilder sfb) {
		boolean hasApplied = false;

		int i = 0;

		// notice that I am using a non-recommended "i--"

		while (i < proofTree.getPBCandidates().size() && !proofTree.isClosed()) {

			// for each signed formula not used, if it accepts a one premise
			// rule, apply the rule and remove it form the list of candidates.

			SignedFormula sf = (SignedFormula) proofTree.getPBCandidates().get(
					i);

			hasApplied = chooseAndApplyOnePremiseRule(proofTree, sfb, sf);
			if (hasApplied) {
				i--;
			}

			i++;
		}

		return hasApplied;
	}

	/**
	 * chooses and applies a one premise rule.
	 * 
	 * @param proofTree
	 * @param sfb
	 * @param hasApplied
	 * @param sf
	 * @return true if the rule has been applied
	 */
	private boolean chooseAndApplyOnePremiseRule(
			ClassicalProofTree proofTree,
			SignedFormulaBuilder sfb, SignedFormula sf) {
		boolean hasApplied = false;
		Rule r = chooseOnePremiseRule(proofTree, sf);

		if (r != NullRule.INSTANCE) {
			hasApplied = true;

			SignedFormulaList sfl = r.getPossibleConclusions(sfb
					.getSignedFormulaFactory(), sfb.getFormulaFactory(),
					new SignedFormulaList(sf));

			proofTree.removeFromPBCandidates(sf,
					SignedFormulaNodeState.ANALYSED);

			for (int j = 0; j < sfl.size(); j++) {
				proofTree.addLast(new SignedFormulaNode(sfl.get(j),
						SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r,
								proofTree.getNode(sf), null)));
			}
		}
		return hasApplied;
	}

	/**
	 * Applies rules with TOP or BOTTOM unary connectives.
	 * 
	 * @param proofTree
	 * @param sfb -
	 *            the formula builder
	 * @return true if some rule was applied.
	 */
	protected boolean applyTOPandBOTTOMrules(
			ClassicalProofTree proofTree,
			SignedFormulaBuilder sfb) {

		Formula topFormula = sfb
				.createCompositeFormula(ClassicalConnectives.TOP);
		Formula bottomFormula = sfb
				.createCompositeFormula(ClassicalConnectives.BOTTOM);
		TopBottomRoleRuleList topAndBottomRules = (TopBottomRoleRuleList) getMethod()
				.getRules().get("topAndBottomRulesNew");
		SignedFormulaList formulasWithTOPorBOTTOM = calculate(proofTree,
				topFormula, bottomFormula);

		return chooseAndApplyTBRules(proofTree, sfb, topFormula, bottomFormula,
				topAndBottomRules, formulasWithTOPorBOTTOM);

	}

	FormulaList getSubformulaLocalReferences(
			ClassicalProofTree proofTree, Formula topFormula,
			SignedFormula aSignedFormula) {
		return ((FormulaReferenceClassicalProofTree) proofTree).getSubformulaLocalReferences(topFormula,
				aSignedFormula);
	}

	/**
	 * @param proofTree
	 * @param sfb
	 * @param hasApplied
	 * @param topFormula
	 * @param bottomFormula
	 * @param topAndBottomRules
	 * @param formulasWithTOPorBOTTOM
	 * @return
	 */
	private boolean chooseAndApplyTBRules(
			ClassicalProofTree proofTree,
			SignedFormulaBuilder sfb, Formula topFormula,
			Formula bottomFormula, TopBottomRoleRuleList topAndBottomRules,
			SignedFormulaList formulasWithTOPorBOTTOM) {

		boolean hasApplied = false;

		while (formulasWithTOPorBOTTOM.size() > 0 && !proofTree.isClosed()) {
			SignedFormula aSignedFormulaWithTOPorBOTTOM = formulasWithTOPorBOTTOM
					.get(0);
			FormulaList fl = getSubformulaLocalReferences(proofTree,
					topFormula, aSignedFormulaWithTOPorBOTTOM);
			fl.addAll(getSubformulaLocalReferences(proofTree, bottomFormula,
					aSignedFormulaWithTOPorBOTTOM));
			CompositeFormula cf1 = (CompositeFormula) fl.get(0);

			CompositeFormula cf1_left = null, cf1_right = null;
			if (cf1.getImmediateSubformulas().get(0) instanceof CompositeFormula) {
				cf1_left = (CompositeFormula) cf1.getImmediateSubformulas()
						.get(0);
			}
			if (cf1.getImmediateSubformulas().size() > 1
					&& cf1.getImmediateSubformulas().get(1) instanceof CompositeFormula) {
				cf1_right = (CompositeFormula) cf1.getImmediateSubformulas()
						.get(1);
			}

			Rule r = null;
			SignedFormula sf = null;

			if (cf1_left == topFormula || cf1_left == bottomFormula) {
				r = topAndBottomRules.get(cf1_left.getConnective(), cf1
						.getConnective(), KERuleRole.LEFT);
				sf = getSignedFormula(r, sfb, aSignedFormulaWithTOPorBOTTOM);
			} else {
				if (cf1_right == topFormula || cf1_right == bottomFormula) {
					r = topAndBottomRules.get(cf1_right.getConnective(), cf1
							.getConnective(), KERuleRole.RIGHT);
					sf = getSignedFormula(r, sfb, aSignedFormulaWithTOPorBOTTOM);
				}
			}

			formulasWithTOPorBOTTOM.remove(aSignedFormulaWithTOPorBOTTOM);
			proofTree.removeFromPBCandidates(aSignedFormulaWithTOPorBOTTOM,
					SignedFormulaNodeState.ANALYSED);
			proofTree.addLast(new SignedFormulaNode(sf,
					SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r,
							proofTree.getNode(aSignedFormulaWithTOPorBOTTOM),
							null)));

			hasApplied = true;

			if (topFormula.isStrictSubformula(sf.getFormula())
					|| bottomFormula.isStrictSubformula(sf.getFormula())) {
				if (!formulasWithTOPorBOTTOM.contains(sf)) {
					formulasWithTOPorBOTTOM.add(0, sf);
				}
			}
		}
		return hasApplied;
	}

	/**
	 * gets the conclusion of a rule.
	 * 
	 * @param r -
	 *            the rule
	 * @param sfb -
	 *            the formula builder
	 * @param aFormula -
	 *            the signed formula
	 * @return the conclusion
	 */
	private SignedFormula getSignedFormula(Rule r, SignedFormulaBuilder sfb,
			SignedFormula aFormula) {
		SignedFormulaList sfl = r.getPossibleConclusions(sfb
				.getSignedFormulaFactory(), sfb.getFormulaFactory(),
				new SignedFormulaList(aFormula));
		if (sfl.size() > 0) {
			return sfl.get(0);
		} else
			throw new TPException("Rule applied but no conclusion: " + r);
	}

	SignedFormulaList getParentReferences(
			ClassicalProofTree proofTree, Formula f) {
		return ((FormulaReferenceClassicalProofTree) proofTree).getParentReferences(f);
	}

	FormulaList getSubformulaParentReferences(
			ClassicalProofTree proofTree, Formula f,
			SignedFormula sf) {
		return ((FormulaReferenceClassicalProofTree) proofTree).getSubformulaParentReferences(f, sf);
	}

	/**
	 * Applies as much two premise rules as possible.
	 * 
	 * @param proofTree -
	 *            the current proof tree
	 * @param sfb -
	 *            the formula builder
	 * @return true if any rule has been applied
	 */
	protected boolean applyTwoPremiseRulesWithMap(
			ClassicalProofTree proofTree,
			SignedFormulaBuilder sfb) {

		boolean hasApplied = false;
		// RuleList twoPremiseRules = getMethod().getRules()
		// .get("twoPremiseRulesNew");

		ConnectiveRoleSignRuleList twoPremiseRulesII = (ConnectiveRoleSignRuleList) getMethod()
				.getRules().get("twoPremiseRulesNewII");

		Formula topFormula = sfb
				.createCompositeFormula(ClassicalConnectives.TOP);
		Formula bottomFormula = sfb
				.createCompositeFormula(ClassicalConnectives.BOTTOM);

		SignedFormula T_TOP = sfb.createSignedFormula(ClassicalSigns.TRUE,
				topFormula);
		SignedFormula F_BOT = sfb.createSignedFormula(ClassicalSigns.FALSE,
				bottomFormula);

		ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
				.getGlobalIterator(), firstInTheBranch = (ProofTreeGlobalIterator) proofTree
				.getGlobalIterator();

		while (!proofTree.isClosed()) {

			SignedFormulaNode sfn = null;
			if (it.hasNextLeft()) {
				sfn = (SignedFormulaNode) it.nextRight();
			} else if (firstInTheBranch.hasPrevious()) {
				sfn = (SignedFormulaNode) firstInTheBranch.previous();
			} else
				break;

			// if it is T_TOP or F_BOT, forget

			if (((SignedFormula) sfn.getContent() == T_TOP)
					|| ((SignedFormula) sfn.getContent() == F_BOT)) {
				continue;
			}

			SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
			SignedFormulaList mainCandidates = intersect(proofTree
					.getPBCandidates(), getParentReferences(proofTree,
					auxCandidate.getFormula()));
			mainCandidates.remove(auxCandidate);

			for (int candidate = 0; candidate < mainCandidates.size(); candidate++) {
				SignedFormula mainCandidate = mainCandidates.get(candidate);

				// for (int i = 0; i < twoPremiseRules.size(); i++) {
				// TwoPremisesOneConclusionRule r =
				// (TwoPremisesOneConclusionRule) twoPremiseRules
				// .get(i);

				SignedFormulaList sflToApply = new SignedFormulaList(
						mainCandidate);
				sflToApply.add(auxCandidate);

				FormulaList fl = getSubformulaParentReferences(proofTree,
						auxCandidate.getFormula(), mainCandidate);
				if (fl.size() >= 1) {
					// System.err.println (mainCandidate + " " + auxCandidate );
					CompositeFormula f = (CompositeFormula) fl.get(0);
					Connective conn = f.getConnective();
					// System.err.println(fl + " " + conn);

					Formula f_left = (Formula) f.getImmediateSubformulas().get(
							0);
					Formula f_right = null;
					if (f.getImmediateSubformulas().size() > 1) {
						f_right = (Formula) f.getImmediateSubformulas().get(1);
					}
					KERuleRole role = (auxCandidate.getFormula() == f_left) ? KERuleRole.LEFT
							: ((f_right != null) && (auxCandidate.getFormula() == f_right)) ? KERuleRole.RIGHT
									: null;
					FormulaSign sign = auxCandidate.getSign();

					// System.err.println(role);

					TwoPremisesOneConclusionRule r1 = (TwoPremisesOneConclusionRule) twoPremiseRulesII
							.get(conn, role, sign);
					// System.err.println(r1);

					SignedFormulaList sflResult = r1.getPossibleConclusions(sfb
							.getSignedFormulaFactory(),
							sfb.getFormulaFactory(), sflToApply, f);
					if (sflResult != null) {
						SignedFormula conclusion = sflResult.get(0);

						SignedFormulaNode n;
						proofTree.addLast(n = new SignedFormulaNode(conclusion,
								SignedFormulaNodeState.NOT_ANALYSED,
								createOrigin(r1, proofTree
										.getNode(mainCandidate), proofTree
										.getNode(auxCandidate))));

						proofTree.removeFromPBCandidates(mainCandidate,
								SignedFormulaNodeState.ANALYSED);

						hasApplied = true;
					}

				}

				// }
			}
		}

		return hasApplied;
	}

	/**
	 * applies the PB rule
	 * 
	 * @param s
	 * @param sfb
	 * @return
	 */
	protected boolean applyPBRule(LinkedList s, SignedFormulaBuilder sfb) {

		boolean hasApplied = false;

		SignedFormulaList candidates = getThisCurrent().getPBCandidates();

		if (candidates.size() > 0) {
			return applyPBOnce(s, sfb, candidates);

		}

		return hasApplied;
	}

	/**
	 * @param s
	 * @param sfb
	 * @param PBRules
	 * @param candidates
	 * @return
	 */
	private boolean applyPBOnce(LinkedList s, SignedFormulaBuilder sfb,
			SignedFormulaList candidates) {
		PBRuleList PBRules = (PBRuleList) getMethod().getRules().get("PBRules");
		SignedFormula candidateChosen = candidates.get(0);

		// System.err.println(candidateChosen);
		TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule) PBRules
				.get(candidateChosen.getSign(),
						((CompositeFormula) candidateChosen.getFormula())
								.getConnective());

		SignedFormula aux = r.getAuxiliaryCandidates(
				sfb.getSignedFormulaFactory(), sfb.getFormulaFactory(),
				candidateChosen).get(0);

		SignedFormula auxOpposite = (aux.getSign().equals(ClassicalSigns.TRUE)) ? sfb
				.createSignedFormula(ClassicalSigns.FALSE, aux.getFormula())
				: sfb
						.createSignedFormula(ClassicalSigns.TRUE, aux
								.getFormula());

		// TODO verificar se concluso j est na ProofTree
		SignedFormulaList sfl = new SignedFormulaList(candidateChosen);
		sfl.add(aux);
		SignedFormula conclusion = r.getPossibleConclusions(
				sfb.getSignedFormulaFactory(), sfb.getFormulaFactory(), sfl)
				.get(0);
		getThisCurrent().removeFromPBCandidates(candidateChosen,
				SignedFormulaNodeState.ANALYSED);

		FormulaReferenceClassicalProofTree right = (FormulaReferenceClassicalProofTree) getThisCurrent()
				.addRight(
						new SignedFormulaNode(auxOpposite,
								SignedFormulaNodeState.NOT_ANALYSED,
								createOrigin(ClassicalRules.PB,
										getThisCurrent().getNode(
												candidateChosen), null)));

		Rule complementary = PBRules.getComplementary(r);
		if (complementary != null) {
			SignedFormulaList sflComplementary = new SignedFormulaList(
					candidateChosen);
			sflComplementary.add(auxOpposite);
			SignedFormula complementaryConclusion = complementary
					.getPossibleConclusions(sfb.getSignedFormulaFactory(),
							sfb.getFormulaFactory(), sflComplementary).get(0);

			right.addLast(new SignedFormulaNode(complementaryConclusion,
					SignedFormulaNodeState.NOT_ANALYSED, createOrigin(
							complementary, getThisCurrent().getNode(
									candidateChosen), right
									.getNode(auxOpposite))));
		}

		FormulaReferenceClassicalProofTree left = (FormulaReferenceClassicalProofTree) getThisCurrent()
				.addLeft(
						new SignedFormulaNode(aux,
								SignedFormulaNodeState.NOT_ANALYSED,
								createOrigin(ClassicalRules.PB,
										getThisCurrent().getNode(
												candidateChosen), null)));
		left.addLast(new SignedFormulaNode(conclusion,
				SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r,
						getThisCurrent().getNode(candidateChosen), left
								.getNode(aux))));
		//         s.addFirst(_current);
		s.addFirst(right);
		setCurrent(left);

		return true;
	}

	/**
	 * @param auxCandidate
	 * @param mainCandidate
	 * @return
	 */
	private KERuleRole findRole(SignedFormula auxCandidate,
			SignedFormula mainCandidate) {

		if (KERuleRole.LEFT.getFormulas(mainCandidate.getFormula()).contains(
				auxCandidate.getFormula())) {
			return KERuleRole.LEFT;
		} else if (KERuleRole.RIGHT.getFormulas(mainCandidate.getFormula())
				.contains(auxCandidate.getFormula())) {
			return KERuleRole.RIGHT;
		}

		return null;
	}

	/**
	 * @param fcpt
	 * @param sfb
	 * @param r
	 * @param sf
	 * @param sfl_left
	 */
	private boolean applyEach(FormulaReferenceClassicalProofTree fcpt,
			SignedFormulaBuilder sfb, TwoPremisesOneConclusionRule r,
			SignedFormula sf, SignedFormulaList sfl) {

		boolean hasApplied = false;
		SignedFormulaList toApply;

		for (int i = 0; i < sfl.size(); i++) {
			toApply = new SignedFormulaList();
			toApply.add(sf);
			toApply.add(sfl.get(i));

			SignedFormulaList conclusions = r.getPossibleConclusions(sfb
					.getSignedFormulaFactory(), sfb.getFormulaFactory(),
					toApply);
			System.out.println("conclusions: " + "  " + r + "  " + toApply
					+ ":" + conclusions);
			if (conclusions != null) {
				hasApplied = true;
				SignedFormula conclusion = conclusions.get(0);

				// coloca a nova na proof tree
				fcpt.addLast(new SignedFormulaNode(conclusion,
						SignedFormulaNodeState.NOT_ANALYSED, createOrigin(r,
								fcpt.getNode(sf), fcpt.getNode(sfl.get(i)))));

			}

		}

		return hasApplied;

	}

	/**
	 * @param fcpt
	 * @param f_left
	 * @return
	 */
	private SignedFormulaList getOccurrences(
			FormulaReferenceClassicalProofTree fcpt, Formula f) {
		SignedFormulaList sfl = new SignedFormulaList();
		ProofTreeGlobalIterator ptgi = (ProofTreeGlobalIterator) fcpt
				.getGlobalIterator();

		// goes to the last
		while (ptgi.hasNextLeft()) {
			ptgi.nextLeft();
		}

		// and then up
		while (ptgi.hasPrevious()) {
			SignedFormulaNode sfn = (SignedFormulaNode) ptgi.current();
			ptgi.previous();

			if (((SignedFormula) sfn.getContent()).getFormula().equals(f)) {
				sfl.add((SignedFormula) sfn.getContent());
			}
			// System.out.println("vendo " + sfn.getContent());
		}

		return sfl;

	}

	/**
	 * @param cpt
	 * @param topFormula
	 * @param bottomFormula
	 * @return
	 */
	private SignedFormulaList calculate(ClassicalProofTree cpt,
			Formula topFormula, Formula bottomFormula) {
		SignedFormulaList sfl = ((FormulaReferenceClassicalProofTree) cpt).getLocalReferences(topFormula);
		sfl.addAll(((FormulaReferenceClassicalProofTree) cpt).getLocalReferences(bottomFormula));
		sfl = intersect(cpt.getPBCandidates(), sfl);
		return sfl;
	}

	/**
	 * @param candidates
	 * @param sfl
	 * @return
	 */
	private SignedFormulaList intersect(SignedFormulaList candidates,
			SignedFormulaList sfl) {
		SignedFormulaList result = new SignedFormulaList();

		for (int i = 0; i < candidates.size(); i++) {
			if (sfl.contains(candidates.get(i))) {
				result.add(candidates.get(i));
			}
		}

		return result;

	}

	private Rule chooseOnePremiseRule(ClassicalProofTree cpt,
			SignedFormula sf) {

		// System.err.println(sf);

		OnePremiseRuleList onePremiseRules = (OnePremiseRuleList) getMethod()
				.getRules().get("onePremiseRules");

		if (sf.getFormula() instanceof CompositeFormula) {
			return onePremiseRules.get(sf.getSign(), ((CompositeFormula) sf
					.getFormula()).getConnective());
		}

		return NullRule.INSTANCE;

	}

	// /**
	// * Applies as much one premise rules as possible
	// *
	// * @param proofTree -
	// * the current proof tree
	// * @param sfb -
	// * the formula builder
	// * @return true if any rule has been applied
	// */
	// private boolean applyOnePremiseRulesWithIterator(
	// FormulaReferenceClassicalProofTree proofTree,
	// SignedFormulaBuilder sfb) {
	// /*
	// *
	// * iterate over current branch
	// *
	// * for each formula, see if a 1-premise rule is applicable. If it is,
	// * apply it and remove it from list of formulas for which 1-premise
	// * rules can be applied.
	// *
	// */
	// boolean hasApplied = false;
	//
	// Iterator it = proofTree.getPBCandidates().iterator();
	//
	// try {
	// while (it.hasNext()) {
	// SignedFormula sf = (SignedFormula) it.next();
	//
	// hasApplied = chooseAndApplyOnePremiseRule(proofTree, sfb, sf);
	//
	// }
	// } catch (ConcurrentModificationException e) {
	// // No problem. Same thread.
	// // TODO it would be wrong to write a program that depended on this
	// // exception for its correctness: ConcurrentModificationException
	// // should be used only to detect bugs.
	// }
	//
	// return hasApplied;
	// }

	// /**
	// * Applies as much two premise rules as possible.
	// *
	// * @param proofTree -
	// * the current proof tree
	// * @param sfb -
	// * the formula builder
	// * @return true if any rule has been applied
	// */
	// private boolean applyTwoPremiseRules(
	// FormulaReferenceClassicalProofTree proofTree,
	// SignedFormulaBuilder sfb) {
	//
	// boolean hasApplied = false;
	// RuleList twoPremiseRules = getMethod().getRules()
	// .get("twoPremiseRulesNew");
	//
	// ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
	// .getGlobalIterator();
	//
	// // AQUI
	// // TODO Deixar de ir de baixo para cima. Criar novo iterador.
	// // goes to the last
	// while (it.hasNextLeft()) {
	// it.nextLeft();
	// }
	// // ProofTreeBasicIterator it = fcpt.getGlobalIterator();
	// //TODO deixar de olhar o F BOT
	// while (it.hasPrevious() && !proofTree.isClosed()) {
	// SignedFormulaNode sfn = (SignedFormulaNode) it.current();
	// it.previous();
	//
	// // System.err.println(" FAss do n " + sfn.getContent());
	// SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
	// // System.err.println(" Ref  frmula na rvore "
	// // + fcpt.getParentReferences(auxCandidate.getFormula()));
	// SignedFormulaList cand = intersect(proofTree.getPBCandidates(),
	// proofTree.getParentReferences(auxCandidate.getFormula()));
	// cand.remove(auxCandidate);
	// // System.err.println(" Ref cand  frmula na rvore " + cand);
	//
	// for (int candidate = 0; candidate < cand.size(); candidate++) {
	// SignedFormula mainCandidate = cand.get(candidate);
	//
	// // TODO no funciona porque pode ser um conectivo interior
	// // Connective mainConnective = ((CompositeFormula) mainCandidate
	// // .getFormula()).getConnective();
	// // System.err.print(mainConnective + " \n");
	//
	// for (int i = 0; i < twoPremiseRules.size(); i++) {
	// TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule)
	// twoPremiseRules
	// .get(i);
	//
	// // System.err.print(r + " ");
	//
	// SignedFormulaList sflToApply = new SignedFormulaList(
	// mainCandidate);
	// sflToApply.add(auxCandidate);
	// // System.err.print(sflToApply);
	//
	// // TODO no funciona porque pode ser subfrmula
	// // KERuleRole auxRole = findRole(auxCandidate,
	// // mainCandidate);
	//
	// // System.err.print(auxRole + " ");
	//
	// SignedFormulaList sflResult = r.getPossibleConclusions(sfb
	// .getSignedFormulaFactory(),
	// sfb.getFormulaFactory(), sflToApply);
	// // System.err.println(" result " + sflResult);
	// if (sflResult != null) {
	// SignedFormula conclusion = sflResult.get(0);
	//
	// // coloca a nova na proof tree
	// SignedFormulaNode n;
	// proofTree.addLast(n = new SignedFormulaNode(conclusion,
	// SignedFormulaNodeState.NOT_ANALYSED,
	// createOrigin(r, proofTree.getNode(mainCandidate),
	// proofTree.getNode(auxCandidate))));
	//
	// // System.err.println ("Nodeadded " +n + " at " + fcpt);
	//
	// proofTree.removeFromPBCandidates(mainCandidate,
	// SignedFormulaNodeState.ANALYSED);
	//
	// hasApplied = true;
	// }
	//
	// }
	//
	// }
	//
	// }
	//
	// return hasApplied;
	// }

	/**
	 * Applies as much two premise rules as possible.
	 * 
	 * @param proofTree -
	 *            the current proof tree
	 * @param sfb -
	 *            the formula builder
	 * @return true if any rule has been applied
	 */
	private boolean applyTwoPremiseRulesWithoutMapBottomUp(
			FormulaReferenceClassicalProofTree proofTree,
			SignedFormulaBuilder sfb) {

		boolean hasApplied = false;

		RuleList twoPremiseRules = getMethod().getRules().get(
				"twoPremiseRulesNew");

		ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
				.getGlobalIterator();

		Formula topFormula = sfb
				.createCompositeFormula(ClassicalConnectives.TOP);
		Formula bottomFormula = sfb
				.createCompositeFormula(ClassicalConnectives.BOTTOM);

		SignedFormula T_TOP = sfb.createSignedFormula(ClassicalSigns.TRUE,
				topFormula);
		SignedFormula F_BOT = sfb.createSignedFormula(ClassicalSigns.FALSE,
				bottomFormula);

		// AQUI
		// TODO Deixar de ir de baixo para cima. Criar novo iterador.
		// goes to the last
		while (it.hasNextLeft()) {
			it.nextLeft();
		}
		// ProofTreeBasicIterator it = fcpt.getGlobalIterator();
		// TODO deixar de olhar o F BOT
		while (it.hasPrevious() && !proofTree.isClosed()) {
			SignedFormulaNode sfn = (SignedFormulaNode) it.current();
			it.previous();

			if (((SignedFormula) sfn.getContent() == T_TOP)
					|| ((SignedFormula) sfn.getContent() == F_BOT)) {
				continue;
			}

			// System.err.println(" FAss do n " + sfn.getContent());
			SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
			// System.err.println(" Ref  frmula na rvore "
			// + fcpt.getParentReferences(auxCandidate.getFormula()));
			SignedFormulaList cand = intersect(proofTree.getPBCandidates(),
					getParentReferences(proofTree, auxCandidate.getFormula()));
			cand.remove(auxCandidate);
			// System.err.println(" Ref cand  frmula na rvore " + cand);

			for (int candidate = 0; candidate < cand.size(); candidate++) {
				SignedFormula mainCandidate = cand.get(candidate);

				// TODO no funciona porque pode ser um conectivo interior
				// Connective mainConnective = ((CompositeFormula) mainCandidate
				// .getFormula()).getConnective();
				// System.err.print(mainConnective + " \n");

				for (int i = 0; i < twoPremiseRules.size(); i++) {
					TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule) twoPremiseRules
							.get(i);

					// System.err.print(r + " ");

					SignedFormulaList sflToApply = new SignedFormulaList(
							mainCandidate);
					sflToApply.add(auxCandidate);
					// System.err.print(sflToApply);

					// TODO no funciona porque pode ser subfrmula
					// KERuleRole auxRole = findRole(auxCandidate,
					// mainCandidate);

					// System.err.print(auxRole + " ");

					SignedFormulaList sflResult = r.getPossibleConclusions(sfb
							.getSignedFormulaFactory(),
							sfb.getFormulaFactory(), sflToApply);
					// System.err.println(" result " + sflResult);
					if (sflResult != null) {
						SignedFormula conclusion = sflResult.get(0);

						// coloca a nova na proof tree
						SignedFormulaNode n;
						proofTree.addLast(n = new SignedFormulaNode(conclusion,
								SignedFormulaNodeState.NOT_ANALYSED,
								createOrigin(r, proofTree
										.getNode(mainCandidate), proofTree
										.getNode(auxCandidate))));

						// System.err.println ("Nodeadded " +n + " at " + fcpt);

						proofTree.removeFromPBCandidates(mainCandidate,
								SignedFormulaNodeState.ANALYSED);

						hasApplied = true;
					}

				}

			}

		}

		return hasApplied;
	}

	/**
	 * Applies as much two premise rules as possible.
	 * 
	 * @param proofTree -
	 *            the current proof tree
	 * @param sfb -
	 *            the formula builder
	 * @return true if any rule has been applied
	 */
	private boolean applyTwoPremiseRulesWithoutMapCurrentDownThenUp(
			FormulaReferenceClassicalProofTree proofTree,
			SignedFormulaBuilder sfb) {

		// TODO REMOVE
		ConnectiveRoleSignRuleList twoPremiseRulesII = (ConnectiveRoleSignRuleList) getMethod()
				.getRules().get("twoPremiseRulesNewII");

		boolean hasApplied = false;
		RuleList twoPremiseRules = getMethod().getRules().get(
				"twoPremiseRulesNewII");

		ProofTreeGlobalIterator it = (ProofTreeGlobalIterator) proofTree
				.getGlobalIterator(), firstInTheBranch = (ProofTreeGlobalIterator) proofTree
				.getGlobalIterator();

		Formula topFormula = sfb
				.createCompositeFormula(ClassicalConnectives.TOP);
		Formula bottomFormula = sfb
				.createCompositeFormula(ClassicalConnectives.BOTTOM);

		SignedFormula T_TOP = sfb.createSignedFormula(ClassicalSigns.TRUE,
				topFormula);
		SignedFormula F_BOT = sfb.createSignedFormula(ClassicalSigns.FALSE,
				bottomFormula);

		while (!proofTree.isClosed()) {

			SignedFormulaNode sfn = null;
			if (it.hasNextLeft()) {
				sfn = (SignedFormulaNode) it.nextRight();
			} else if (firstInTheBranch.hasPrevious()) {
				sfn = (SignedFormulaNode) firstInTheBranch.previous();
			} else
				break;

			if (((SignedFormula) sfn.getContent() == T_TOP)
					|| ((SignedFormula) sfn.getContent() == F_BOT)) {
				continue;
			}

			SignedFormula auxCandidate = (SignedFormula) sfn.getContent();
			SignedFormulaList mainCandidates = intersect(proofTree
					.getPBCandidates(), getParentReferences(proofTree,
					auxCandidate.getFormula()));
			mainCandidates.remove(auxCandidate);

			for (int candidate = 0; candidate < mainCandidates.size(); candidate++) {
				SignedFormula mainCandidate = mainCandidates.get(candidate);

				for (int i = 0; i < twoPremiseRules.size(); i++) {
					TwoPremisesOneConclusionRule r = (TwoPremisesOneConclusionRule) twoPremiseRules
							.get(i);

					SignedFormulaList sflToApply = new SignedFormulaList(
							mainCandidate);
					sflToApply.add(auxCandidate);

					// FormulaList fl = proofTree.getSubformulaParentReferences(
					// auxCandidate.getFormula(), mainCandidate);
					// if (fl.size() > 1) {
					// // System.err.println (mainCandidate + " " +
					// // auxCandidate );
					// CompositeFormula f = (CompositeFormula) fl.get(0);
					// Connective conn = f.getConnective();
					// // System.err.println(fl + " " + conn);
					//                    
					// Formula f_left = (Formula) f.getImmediateSubformulas()
					// .get(0);
					// Formula f_right = null;
					// if (f.getImmediateSubformulas().size() > 1) {
					// f_right = (Formula) f.getImmediateSubformulas()
					// .get(1);
					// }
					// KERuleRole role = (auxCandidate.getFormula() == f_left) ?
					// KERuleRole.LEFT
					// : ((f_right != null) && (auxCandidate
					// .getFormula() == f_right)) ? KERuleRole.RIGHT
					// : null;
					//                    
					// // System.err.println(role);
					// FormulaSign sign = auxCandidate.getSign();
					//                    
					// TwoPremisesOneConclusionRule r1 =
					// (TwoPremisesOneConclusionRule) twoPremiseRulesII
					// .get(conn, role, sign);
					// // System.err.println(r1);
					//                    
					// SignedFormulaList sflResultX = r1
					// .getPossibleConclusions(sfb
					// .getSignedFormulaFactory(), sfb
					// .getFormulaFactory(), sflToApply, f);
					//
					// // repeteco - incio
					// if (sflResultX != null) {
					// SignedFormula conclusion = sflResultX.get(0);
					//
					// SignedFormulaNode n;
					// proofTree.addLast(n = new SignedFormulaNode(conclusion,
					// SignedFormulaNodeState.NOT_ANALYSED,
					// createOrigin(r1, proofTree.getNode(mainCandidate),
					// proofTree.getNode(auxCandidate))));
					//
					// proofTree.removeFromPBCandidates(mainCandidate,
					// SignedFormulaNodeState.ANALYSED);
					//
					// hasApplied = true;
					// }
					// // repeteco - fim
					// }

					SignedFormulaList sflResult = r.getPossibleConclusions(sfb
							.getSignedFormulaFactory(),
							sfb.getFormulaFactory(), sflToApply);
					if (sflResult != null) {
						SignedFormula conclusion = sflResult.get(0);

						SignedFormulaNode n;
						proofTree.addLast(n = new SignedFormulaNode(conclusion,
								SignedFormulaNodeState.NOT_ANALYSED,
								createOrigin(r, proofTree
										.getNode(mainCandidate), proofTree
										.getNode(auxCandidate))));

						proofTree.removeFromPBCandidates(mainCandidate,
								SignedFormulaNodeState.ANALYSED);

						hasApplied = true;
					}

				}
			}
		}

		return hasApplied;
	}

}