/*
 * Created on 24/10/2003, 13:55:23 
 *
 */
package parsers;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringReader;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;

import JSX.ObjectWriter;

/**
 * A class for using different parsers constructed with java_cup.
 * 
 * @author Adolfo Neto
 */

public class ParserUser {

    ClassLoader _classLoader;

    /**
     * creates a parser user object using the default class loader
     * 
     */
    public ParserUser() {
        _classLoader = this.getClass().getClassLoader();
    }

    /**
     * creates a parser user using the a class loader that uses jar files
     * 
     * @param logicJarFiles
     */
    public ParserUser(String[] jarFiles) {

        try {

            URL[] jarURLs = new URL[jarFiles.length];

            for (int i = 0; i < jarFiles.length; i++) {

                // verifies existence of file
                // removes "file:"
                String name = jarFiles[i].substring(5);
                try {
                    FileReader fr = new FileReader(name);
                } catch (Exception e) {
                    System.err.println("ParserUser: File not found: " + name);
                    e.printStackTrace();
                    System.exit(-1);
                }

                jarURLs[i] = new URL(jarFiles[i]);
            }

            _classLoader = new URLClassLoader(jarURLs);

        } catch (MalformedURLException e) {
            System.err.print("ParserUser: ");
            System.err
                    .println("Unable to create parser for at least one of the following packages: "
                            + Arrays.asList(jarFiles));
            e.printStackTrace();
            System.exit(-1);
        }

    }

    public Object parseString(String lexerClassName, String parserClassName,
            String str) {

        Object lexer = null, parser = null;
        try {
            lexer = createLexerForString(lexerClassName, str);
        } catch (Exception e) {
            System.out.println("ParserUser.parseString: didn't find class "
                    + lexerClassName);
            e.printStackTrace();
            System.exit(-1);
        }
        try {
            parser = createParser(parserClassName, lexer);
        } catch (Exception e) {
            System.out.println("ParserUser.parseString: didn't find class "
                    + parserClassName);
            e.printStackTrace();
            System.exit(-1);
        }

        Object result=null;
        try {
            result = ((java_cup.runtime.lr_parser) parser).parse().value;
        } catch (Exception e1) {
            e1.printStackTrace();
            System.exit(-1);
        }
        return result;
    }

    public Object parseFile(String lexerClassName, String parserClassName,
            String filename) {

        // filename = System.getProperty("user.dir") + File.separator +
        // filename;

        try {
            Object lexer = createLexerForFile(lexerClassName, filename);
            Object parser = createParser(parserClassName, lexer);

            Object result = ((java_cup.runtime.lr_parser) parser).parse().value;

            return result;
        } catch (FileNotFoundException e) {
            System.out.println("ParserUser: File not found: " + filename);
            e.printStackTrace();
            System.exit(-1);
        } catch (Exception e) {
            System.out.println("ParserUser.parseFile: didn't find class.");
            e.printStackTrace();
            System.exit(-1);
        }

        return null;
    }

    /**
     * creates a lexer object from the name of the class lexer
     * 
     * @param lexerClassName
     * @param str
     * @return the lexer object or null
     */
    public Object createLexerForString(String lexerClassName, String str) {

        try {
            Class lexerClass = Class
                    .forName(lexerClassName, true, _classLoader);
            // System.out.println(lexerClass);

            Class[] lexerArgsClasses = new Class[] { java.io.Reader.class };
            Object[] lexerArgs = new Object[] { new StringReader(str) };

            Constructor lexerConstructor;
            // showConstructors(lexerClass);

            try {

                lexerConstructor = lexerClass.getConstructor(lexerArgsClasses);

                try {
                    Object object = null;
                    object = lexerConstructor.newInstance(lexerArgs);
                    return object;

                } catch (InstantiationException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (IllegalArgumentException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (InvocationTargetException e) {
                    e.printStackTrace();
                    System.exit(-1);
                }

            } catch (NoSuchMethodException e) {
                e.printStackTrace();
                System.exit(-1);
            }

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }

        return null;

    }

    /**
     * creates a lexer object from the name of the class lexer
     * 
     * @param lexerClassName
     * @param filename
     * @return the lexer object or null
     */
    public Object createLexerForFile(String lexerClassName, String filename) {

        try {
            Class lexerClass = Class
                    .forName(lexerClassName, true, _classLoader);
            // System.out.println(lexerClass);

            Class[] lexerArgsClasses = new Class[] { java.io.Reader.class };
            Object[] lexerArgs = new Object[] { new FileReader(filename) };
            Constructor lexerConstructor;
            // showConstructors(lexerClass);

            try {

                lexerConstructor = lexerClass.getConstructor(lexerArgsClasses);

                try {
                    Object object = null;
                    object = lexerConstructor.newInstance(lexerArgs);
                    return object;

                } catch (InstantiationException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (IllegalArgumentException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (InvocationTargetException e) {
                    e.printStackTrace();
                    System.exit(-1);
                }

            } catch (NoSuchMethodException e) {
                e.printStackTrace();
                System.exit(-1);
            }

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
        ;

        return null;

    }

    /**
     * creates a parse object from the name of the class parser
     * 
     * @param parserClassName
     * @param str
     * @return the parser object or null
     */
    public Object createParser(String parserClassName, Object lexer) {

        try {
            Class parserClass = Class.forName(parserClassName, true,
                    _classLoader);
            // System.out.println(parserClass);

            Class[] parserArgsClasses = new Class[] { java_cup.runtime.Scanner.class };
            Object[] parserArgs = new Object[] { lexer };
            Constructor parserConstructor;
            // showConstructors(parserClass);

            try {

                parserConstructor = parserClass
                        .getConstructor(parserArgsClasses);

                try {
                    Object object = null;
                    object = parserConstructor.newInstance(parserArgs);

                    return object;

                } catch (InstantiationException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (IllegalAccessException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (IllegalArgumentException e) {
                    e.printStackTrace();
                    System.exit(-1);
                } catch (InvocationTargetException e) {
                    e.printStackTrace();
                    System.exit(-1);
                }

            } catch (NoSuchMethodException e) {
                e.printStackTrace();
                System.exit(-1);
            }

        } catch (Exception e) {
            e.printStackTrace();
            System.exit(-1);
        }
        ;

        return null;

    }

    static void showConstructors(Class c) {

        System.out.println("SHOW CONSTRUCTORS FOR " + c.toString());

        Constructor[] theConstructors = c.getConstructors();

        System.out.println("number: " + theConstructors.length);
        for (int i = 0; i < theConstructors.length; i++) {
            System.out.print("( ");
            Class[] parameterTypes = theConstructors[i].getParameterTypes();
            for (int k = 0; k < parameterTypes.length; k++) {
                String parameterString = parameterTypes[k].getName();
                System.out.print(parameterString + " ");
            }
            System.out.println(")");
        }
    }

    /**
     * Serializes an object with JSX and puts it in a file
     * 
     * @param Object
     *                   o
     * @param String
     *                   filename
     * 
     */
    public void serialize(Object o, String filename) {
        try {
            ObjectWriter out = new ObjectWriter(new FileWriter(filename));
            out.writeObject(o);
            out.close();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * @param args
     *                   <logic.jar>- jar file containing the classes describing a
     *                   logic <parser.jar>- jar file containing the classes for the
     *                   parser <LexerClassName>- name of the lexer class
     *                   <ParserClassName>- name of the parser class <inputFilename>-
     *                   name of the file to be parsed <outputFilename>- name of the
     *                   XML file (serialization of an object made by JSX) to be
     *                   generated
     */
    public static void main(String[] args) {

        if (args.length != 6) {
            System.out
                    .println("Usage: java -jar ParserUser.jar <logic.jar> <parser.jar> <LexerClassName>"
                            + " <ParserClassName> <inputFilename> <outputFilename>");
            System.out
                    .println("<logic.jar> - jar file containing the classes describing a logic\n"
                            + " <parser.jar> - jar file containing the classes for the parser\n"
                            + " <LexerClassName> - name of the lexer class\n"
                            + " <ParserClassName> - name of the parser class\n"
                            + " <inputFilename> - name of the file to be parsed\n"
                            + " <outputFilename> - name of the XML file (serialization of an object made by JSX) to be generated\n");
        } else {

            String[] jarFiles = new String[2];
            jarFiles[0] = args[0];
            jarFiles[1] = args[1];
            ParserUser p = new ParserUser(jarFiles);

            Object o = p.parseFile(args[2], args[3], args[4]);

            p.serialize(o, args[5]);
        }

    }

}